# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

"""
This is an internal PyTango module.
"""

__all__ = ("connection_init",)

__docformat__ = "restructuredtext"


import collections.abc

from tango._tango import (
    Connection,
    DeviceData,
    ExtractAs,
)
from tango.utils import _trace_client, _get_command_inout_param
from tango.utils import _get_new_CallbackAutoDie
from tango.green import green


def __CallBack__cmd_ended_aux(self, fn):
    def __new_fn(cmd_done_event):
        try:
            cmd_done_event.argout = cmd_done_event.argout_raw.extract(
                self.defaultCommandExtractAs
            )
        except Exception:
            pass
        return fn(cmd_done_event)

    return __new_fn


def __Connection__command_inout(self, name, cmd_param=None):
    """
    command_inout( self, cmd_name, cmd_param=None, __GREEN_KWARGS__) -> any

        Execute a command on a device.

    :param cmd_name: Command name
    :type value: str

    :param cmd_param: It should be a value of the type expected by the command or a DeviceData object with this value inserted.
                      It can be omitted if the command should not get any argument.
    :type cmd_param: Any

    __GREEN_KWARGS_DESCRIPTION__

    :returns: The result of the command. The type depends on the command. It may be None
    :rtype: Any

    :throws: TypeError: if cmd_param's type is not compatible with the command
    :throws: :obj:`tango.ConnectionFailed`: Raised in case of a connection failure.
    :throws: :obj:`tango.CommunicationFailed`: Raised in case of a communication failure.
    :throws: :obj:`tango.DevFailed`: Raised in case of a device failure.
    :throws: :obj:`tango.DeviceUnlocked`: Raised in case of a device failure.
    __GREEN_RAISES__

    .. versionadded:: 8.1.0
        *green_mode* parameter.
        *wait* parameter.
        *timeout* parameter.

    .. versionchanged:: 10.0.0
        TypeError's for invalid command input arguments are now more detailed.
        For commands with a DEV_STRING input argument, invalid data will now raise TypeError instead of SystemError.
    """
    r = Connection.command_inout_raw(self, name, cmd_param)
    if isinstance(r, DeviceData):
        try:
            return r.extract(self.defaultCommandExtractAs)
        except Exception:
            return None
    else:
        return r


__Connection__command_inout.__name__ = "command_inout"


def __Connection__command_inout_raw(self, cmd_name, cmd_param=None):
    """
    command_inout_raw( self, cmd_name, cmd_param=None) -> DeviceData

        Execute a command on a device. Does not convert result.

    :param cmd_name: Command name
    :type value: str

    :param cmd_param: It should be a value of the type expected by the command or a DeviceData object with this value inserted.
                      It can be omitted if the command should not get any argument.
    :type cmd_param: Any

    :returns: The result of the command. The type depends on the command. It may be None
    :rtype: Any

    :throws: TypeError: if cmd_param's type is not compatible with the command
    :throws: :obj:`tango.ConnectionFailed`: Raised in case of a connection failure.
    :throws: :obj:`tango.CommunicationFailed`: Raised in case of a communication failure.
    :throws: :obj:`tango.DevFailed`: Raised in case of a device failure.
    :throws: :obj:`tango.DeviceUnlocked`: Raised in case of a device failure.

    .. versionchanged:: 10.0.0
        TypeError's for invalid command input arguments are now more detailed.
        For commands with a DEV_STRING input argument, invalid data will now raise TypeError instead of SystemError.
    """
    param = _get_command_inout_param(self, cmd_name, cmd_param)
    return self.__command_inout(cmd_name, param)


def __Connection__command_inout_asynch(self, cmd_name, *args):
    """
    command_inout_asynch(self, cmd_name) -> id
    command_inout_asynch(self, cmd_name, cmd_param) -> id
    command_inout_asynch(self, cmd_name, cmd_param, forget) -> id

            Execute asynchronously (polling model) a command on a device

        Parameters :
                - cmd_name  : (str) Command name.
                - cmd_param : (any) It should be a value of the type expected by the
                              command or a DeviceData object with this value inserted.
                              It can be omitted if the command should not get any argument.
                              If the command should get no argument and you want
                              to set the 'forget' param, use None for cmd_param.
                - forget    : (bool) If this flag is set to true, this means that the client
                              does not care at all about the server answer and will even
                              not try to get it. Default value is False. Please,
                              note that device re-connection will not take place (in case
                              it is needed) if the fire and forget mode is used. Therefore,
                              an application using only fire and forget requests is not able
                              to automatically re-connnect to device.
        Return     : (int) This call returns an asynchronous call identifier which is
                     needed to get the command result (see command_inout_reply)

        Throws     : ConnectionFailed, TypeError, anything thrown by command_query

    command_inout_asynch( self, cmd_name, callback) -> None
    command_inout_asynch( self, cmd_name, cmd_param, callback) -> None

            Execute asynchronously (callback model) a command on a device.

        Parameters :
                - cmd_name  : (str) Command name.
                - cmd_param : (any)It should be a value of the type expected by the
                              command or a DeviceData object with this value inserted.
                              It can be omitted if the command should not get any argument.
                - callback  : Any callable object (function, lambda...) or any oject
                              with a method named "cmd_ended".
        Return     : None

        Throws     : ConnectionFailed, TypeError, anything thrown by command_query

    .. important::
        by default, TANGO is initialized with the **polling** model. If you want
        to use the **push** model (the one with the callback parameter), you
        need to change the global TANGO model to PUSH_CALLBACK.
        You can do this with the :meth:`tango.ApiUtil.set_asynch_cb_sub_model`

    .. important::
        Multiple asynchronous calls are not guaranteed to be executed by the device
        server in the same order they are invoked by the client.  E.g., a call
        to ``command_inout_asynch("A")`` followed immediately with a call to
        ``command_inout_asynch("B")`` could result in the device invoking
        command ``B`` before command ``A``.

    .. versionchanged:: 10.0.0
        TypeError's for invalid command input arguments are now more detailed.
        For commands with a DEV_STRING input argument, invalid data will now raise TypeError instead of SystemError.
    """
    if len(args) == 0:  # command_inout_asynch()
        argin = DeviceData()
        forget = False
        return self.__command_inout_asynch_id(cmd_name, argin, forget)
    elif len(args) == 1:
        if isinstance(
            args[0], collections.abc.Callable
        ):  # command_inout_asynch(lambda)
            cb = _get_new_CallbackAutoDie()
            cb.cmd_ended = args[0]
            argin = _get_command_inout_param(self, cmd_name)
            return self.__command_inout_asynch_cb(cmd_name, argin, cb)
        elif hasattr(args[0], "cmd_ended"):  # command_inout_asynch(Cbclass)
            cb = _get_new_CallbackAutoDie()
            cb.cmd_ended = args[0].cmd_ended
            argin = _get_command_inout_param(self, cmd_name)
            return self.__command_inout_asynch_cb(cmd_name, argin, cb)
        else:  # command_inout_asynch(value)
            argin = _get_command_inout_param(self, cmd_name, args[0])
            forget = False
            return self.__command_inout_asynch_id(cmd_name, argin, forget)
    elif len(args) == 2:
        if isinstance(
            args[1], collections.abc.Callable
        ):  # command_inout_asynch( value, lambda)
            cb = _get_new_CallbackAutoDie()
            cb.cmd_ended = args[1]
            argin = _get_command_inout_param(self, cmd_name, args[0])
            return self.__command_inout_asynch_cb(cmd_name, argin, cb)
        elif hasattr(args[1], "cmd_ended"):  # command_inout_asynch(value, cbClass)
            cb = _get_new_CallbackAutoDie()
            cb.cmd_ended = args[1].cmd_ended
            argin = _get_command_inout_param(self, cmd_name, args[0])
            return self.__command_inout_asynch_cb(cmd_name, argin, cb)
        else:  # command_inout_asynch(value, forget)
            argin = _get_command_inout_param(self, cmd_name, args[0])
            forget = bool(args[1])
            return self.__command_inout_asynch_id(cmd_name, argin, forget)
    else:
        raise TypeError("Wrong number of attributes!")


__Connection__command_inout_asynch.__name__ = "command_inout_asynch"


def __Connection__command_inout_reply(self, idx, timeout=None):
    """
    command_inout_reply(self, idx, timeout=None) -> DeviceData

            Check if the answer of an asynchronous command_inout is arrived
            (polling model). If the reply is arrived and if it is a valid
            reply, it is returned to the caller in a DeviceData object. If
            the reply is an exception, it is re-thrown by this call. If optional
            `timeout` parameter is not provided an exception is also thrown in case
            of the reply is not yet arrived. If `timeout` is provided, the call will
            wait (blocking the process) for the time specified
            in timeout. If after timeout milliseconds, the reply is still
            not there, an exception is thrown. If timeout is set to 0, the
            call waits until the reply arrived.

        Parameters :
            - idx      : (int) Asynchronous call identifier.
            - timeout  : (int) (optional) Milliseconds to wait for the reply.
        Return     : (DeviceData)
        Throws     : AsynCall, AsynReplyNotArrived, CommunicationFailed, DevFailed from device
    """
    if timeout is None:
        r = self.command_inout_reply_raw(idx)
    else:
        r = self.command_inout_reply_raw(idx, timeout)

    if isinstance(r, DeviceData):
        try:
            return r.extract(self.defaultCommandExtractAs)
        except Exception:
            return None
    else:
        return r


__Connection__command_inout_reply.__name__ = "command_inout_reply"


def connection_init():
    Connection.defaultCommandExtractAs = ExtractAs.Numpy
    Connection.command_inout_raw = __Connection__command_inout_raw
    Connection.command_inout = green(
        _trace_client(__Connection__command_inout), update_signature_and_docstring=True
    )
    Connection.command_inout_asynch = _trace_client(__Connection__command_inout_asynch)
    Connection.command_inout_reply = _trace_client(__Connection__command_inout_reply)



#!/usr/bin/env python3
# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

"""
Trivial power supply device with no external connection or behaviour.
"""

from time import sleep

from tango.server import Device, attribute, command


class PowerSupply(Device):
    @attribute(dtype=float)
    def voltage(self):
        return 1.5

    @command
    def calibrate(self):
        sleep(0.1)


if __name__ == "__main__":
    PowerSupply.run_server()



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
"""
Run a Tango device server and do some things that have either caused
memory leaks in the past, or are suspected of causing memory leaks.
"""

import json
import time
import sys

from tango import (
    AttributeProxy,
    AttrQuality,
    AttrWriteType,
    DevFailed,
    DeviceProxy,
    DevState,
    EnsureOmniThread,
    EventType,
    Group,
)
from tango.server import Device, command, attribute
from tango.test_context import DeviceTestContext
from tango.test_utils import (
    COMMAND_TYPED_VALUES,
    GENERAL_TYPED_VALUES,
    IMAGE_TYPED_VALUES,
    repr_type,
)


class MemTestDevice(Device):
    _attr_data = {}
    _attr_quality = AttrQuality.ATTR_VALID

    def initialize_dynamic_attributes(self):
        for dtype, values in {**GENERAL_TYPED_VALUES, **IMAGE_TYPED_VALUES}.items():
            name = f"attr_ro_{repr_type(dtype)}"
            self._attr_data[name] = {"value": values[0], "reads": 0, "writes": 0}

            attr = attribute(
                name=name,
                dtype=dtype,
                max_dim_x=3,
                max_dim_y=3,
                access=AttrWriteType.READ,
                fget=self.generic_read,
            )
            self.add_attribute(attr)
            self.set_change_event(name, implemented=True, detect=False)

            name = f"attr_rw_{repr_type(dtype)}"
            self._attr_data[name] = {"value": values[0], "reads": 0, "writes": 0}

            attr = attribute(
                name=name,
                dtype=dtype,
                max_dim_x=3,
                max_dim_y=3,
                access=AttrWriteType.READ_WRITE,
                fget=self.generic_read,
                fset=self.generic_write,
            )
            self.add_attribute(attr)
            self.set_change_event(name, implemented=True, detect=False)

        # Include standard attributes
        self.set_change_event("State", implemented=True, detect=False)
        self.set_change_event("Status", implemented=True, detect=False)
        self._attr_data["State"] = {"value": DevState.ON, "reads": 0, "writes": 0}
        self._attr_data["Status"] = {"value": "State ON", "reads": 0, "writes": 0}

    def generic_read(self, attr):
        attr_name = attr.get_name()
        value = self._attr_data[attr_name]["value"]
        self._attr_data[attr_name]["reads"] += 1
        return value, time.time(), self._attr_quality

    def generic_write(self, attr):
        attr_name = attr.get_name()
        value = attr.get_write_value()
        self._attr_data[attr_name]["writes"] += 1
        self._attr_data[attr_name]["value"] = value

    @command
    def cmd_void_void_(self):
        pass

    @command
    def cmd_int_in_(self, value: int):
        pass

    @command
    def cmd_int_out_(self) -> int:
        return 123

    @command
    def cmd_int_in_out_(self, value: int) -> int:
        return value

    @command
    def cmd_float_in_(self, value: float):
        pass

    @command
    def cmd_float_out_(self) -> float:
        return 123.4

    @command
    def cmd_float_in_out_(self, value: float) -> float:
        return value

    @command
    def cmd_str_in_(self, value: str):
        pass

    @command
    def cmd_str_out_(self) -> str:
        return "abc"

    @command
    def cmd_str_in_out_(self, value: str) -> str:
        return value

    @command
    def cmd_bool_in_(self, value: bool):
        pass

    @command
    def cmd_bool_out_(self) -> bool:
        return True

    @command
    def cmd_bool_in_out_(self, value: bool) -> bool:
        return value

    @command
    def cmd_int_list_in_(self, value: list[int]):
        pass

    @command
    def cmd_int_list_out_(self) -> list[int]:
        return [123, 456]

    @command
    def cmd_int_list_in_out_(self, value: list[int]) -> list[int]:
        return value

    @command
    def cmd_float_list_in_(self, value: list[float]):
        pass

    @command
    def cmd_float_list_out_(self) -> list[float]:
        return [123.4, 456.7]

    @command
    def cmd_float_list_in_out_(self, value: list[float]) -> list[float]:
        return value

    @command
    def cmd_str_list_in_(self, value: list[str]):
        pass

    @command
    def cmd_str_list_out_(self) -> list[str]:
        return ["abc", "def"]

    @command
    def cmd_str_list_in_out_(self, value: list[str]) -> list[str]:
        return value

    @command
    def cmd_bool_list_in_(self, value: list[bool]):
        pass

    @command
    def cmd_bool_list_out_(self) -> list[bool]:
        return [True, False]

    @command
    def cmd_bool_list_in_out_(self, value: list[bool]) -> list[bool]:
        return value

    @command
    def cmd_DevVarLongStringArray_in_(self, value: list[list[int], list[str]]):
        pass

    @command
    def cmd_DevVarLongStringArray_out_(self) -> list[list[int], list[str]]:
        return [[123, 456], ["abc", "def"]]

    @command
    def cmd_DevVarLongStringArray_in_out_(
        self, value: list[list[int], list[str]]
    ) -> list[list[int], list[str]]:
        return value

    @command
    def cmd_DevVarDoubleStringArray_in_(self, value: list[list[float], list[str]]):
        pass

    @command
    def cmd_DevVarDoubleStringArray_out_(self) -> list[list[float], list[str]]:
        return [[123.4, 456.7], ["abc", "def"]]

    @command
    def cmd_DevVarDoubleStringArray_in_out_(
        self, value: list[list[float], list[str]]
    ) -> list[list[float], list[str]]:
        return value

    def dev_state(self):
        self._attr_data["State"]["reads"] += 1
        return self._attr_data["State"]["value"]

    def dev_status(self):
        self._attr_data["Status"]["reads"] += 1
        return self._attr_data["Status"]["value"]

    @command
    def set_attr_quality_invalid(self, invalid: bool):
        if invalid:
            self._attr_quality = AttrQuality.ATTR_INVALID
        else:
            self._attr_quality = AttrQuality.ATTR_VALID

    @command
    def emit_events(self, config_json: str):
        config = json.loads(config_json)
        attr_name = config["name"]
        push_count = config["count"]
        value = self._attr_data[attr_name]["value"]
        for _ in range(push_count):
            self.push_change_event(attr_name, value)

    @command
    def get_attr_stats(self, attr_name: str) -> str:
        stats = {
            "reads": self._attr_data[attr_name]["reads"],
            "writes": self._attr_data[attr_name]["writes"],
        }
        return json.dumps(stats)


def get_command_input_data_map():
    result = {}
    for dtype, values in {**GENERAL_TYPED_VALUES, **COMMAND_TYPED_VALUES}.items():
        if not isinstance(dtype, tuple):
            type_name = repr_type(dtype)
        else:
            type_name = f"{repr_type(dtype[0])}_list"
        in_cmd_name = f"cmd_{type_name}_in_"
        in_out_cmd_name = f"cmd_{type_name}_in_out_"
        result[in_cmd_name] = values[0]
        result[in_out_cmd_name] = values[-1]
    return result


def progress_bar(current, total, width=40):
    progress = int(width * current / total)
    bar = "[" + "#" * progress + "-" * (width - progress) + "]"
    percent = f"{(current / total) * 100:.1f}%"
    sys.stdout.write(f"\r{bar} {percent}")
    sys.stdout.flush()


if __name__ == "__main__":
    print("Running PyTango MemTestDevice")
    with DeviceTestContext(MemTestDevice, process=True) as proxy:
        print(f"Device info: {proxy.info()}")

        cmd_names = [cmd for cmd in proxy.get_command_list() if cmd.startswith("cmd_")]
        cmd_input_data_map = get_command_input_data_map()
        print("\nExercising commands:")
        unique_num_cmd = 4
        for cmd_name in cmd_names:
            # use unique numbers, to help isolate causes of leaks
            unique_num_cmd += 1
            for _ in range(unique_num_cmd):
                args = []
                if "_in_" in cmd_name:
                    args = [cmd_input_data_map[cmd_name]]
                proxy.command_inout(cmd_name, *args)
            print(f"{cmd_name:>35} | calls: {unique_num_cmd:>3} |")

        last_unsubscription_time = 0.0
        attr_names = list(proxy.get_attribute_list())
        print("\nExercising attributes:")
        for count, attr_name in enumerate(attr_names):
            # use unique numbers, to help isolate causes of leaks
            offset = count * 10 + unique_num_cmd
            unique_num_ro_invalid = 10 + offset
            unique_num_ro_valid = 12 + offset
            unique_num_rw = 14 + offset if "_rw_" in attr_name else 0
            unique_num_event_no_sub = 16 + offset
            unique_num_event_with_sub = 18 + offset

            # read (invalid quality)
            proxy.set_attr_quality_invalid(True)
            for _ in range(unique_num_ro_invalid):
                value = proxy.read_attribute(attr_name).value

            # read/write (valid quality)
            proxy.set_attr_quality_invalid(False)
            for _ in range(unique_num_ro_valid):
                value = proxy.read_attribute(attr_name).value

            for _ in range(unique_num_rw):
                proxy.write_attribute(attr_name, value)

            config = {"name": attr_name, "count": unique_num_event_no_sub}
            proxy.emit_events(json.dumps(config))

            config = {"name": attr_name, "count": unique_num_event_with_sub}
            eid = proxy.subscribe_event(
                attr_name, EventType.CHANGE_EVENT, lambda x: None
            )
            proxy.emit_events(json.dumps(config))
            proxy.unsubscribe_event(eid)
            last_unsubscription_time = time.time()

            attr_stats = json.loads(proxy.get_attr_stats(attr_name))
            print(
                f"{attr_name:>25} | "
                f"reads: {unique_num_ro_invalid:>3} direct invalid,"
                f" {unique_num_ro_valid:>3} direct valid"
                f" ({attr_stats['reads']:>3} total) | "
                f"writes: {unique_num_rw:>3} direct ({attr_stats['writes']:>3} total) | "
                f"events: {unique_num_event_no_sub:>3} no subscriber,"
                f" {unique_num_event_with_sub:>3} with subscriber,"
                f" ({unique_num_event_no_sub + unique_num_event_with_sub:>3} total)"
            )

        print("\nExercising client creation:")
        device_name = proxy.dev_name()
        offset = (len(attr_names) * 10 + unique_num_cmd + 10) * 2
        num_device_proxy_from_test_context = 2  # server + device
        unique_num_device_proxy = offset

        print("Creating DeviceProxies:")
        n_device_proxies = unique_num_device_proxy - num_device_proxy_from_test_context
        for ind in range(n_device_proxies):
            DeviceProxy(device_name)
            progress_bar(ind, n_device_proxies)
        print(f"\n              DeviceProxy | created: {unique_num_device_proxy:>3} |")

        print("Creating DeviceProxies to non-existent devices:")
        unique_num_device_proxy_unreachable = unique_num_device_proxy + 1
        for ind in range(unique_num_device_proxy_unreachable):
            try:
                DeviceProxy("tango://127.0.0.1:12345/nonexistent/tango/name")
            except DevFailed:
                pass
            progress_bar(ind, unique_num_device_proxy_unreachable)
        print(
            f"\n  unreachable DeviceProxy | created: {unique_num_device_proxy_unreachable:>3} |"
        )

        print("Creating AttributeProxies:")
        unique_num_attr_proxy = unique_num_device_proxy_unreachable + 1
        for ind in range(unique_num_attr_proxy):
            AttributeProxy(f"{device_name}/state")
            progress_bar(ind, unique_num_attr_proxy)
        print(f"\n           AttributeProxy | created: {unique_num_attr_proxy:>3} |")

        print("Creating Groups:")
        unique_num_group = unique_num_attr_proxy + 1
        for ind in range(unique_num_group):
            group = Group("test")
            group.add(device_name)
            progress_bar(ind, unique_num_group)
        print(f"\n                    Group | created: {unique_num_group:>3} |")

        print("Creating EnsureOmniThreads:")
        unique_num_ensure = unique_num_group + 1
        for _ in range(unique_num_ensure):
            with EnsureOmniThread():
                pass
        print(f"\n         EnsureOmniThread | created: {unique_num_ensure:>3} |")

    print("\nMemTestDevice done.")



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
import time

from tango import (
    EventType,
    AttrWriteType,
    AttrConfEventData,
    DevFailed,
    DevError,
)

from tango.server import Device, attribute, command

from tango.test_utils import DeviceTestContext
from tango.utils import EventCallback

A_BIT = 0.5


class EchoDevice(Device):

    scalar_int_value = 1

    @attribute(access=AttrWriteType.READ_WRITE)
    def scalar_int(self) -> int:
        return self.scalar_int_value

    def write_scalar_int(self, new_value):
        self.scalar_int_value = new_value

    @command(dtype_in=[int])
    def change_min_max(self, values):
        w_attr = self.get_device_attr().get_w_attr_by_name("scalar_int")

        w_attr.set_min_value(values[0])
        w_attr.set_max_value(values[1])


def test_attribute_configuration_event():

    with DeviceTestContext(EchoDevice, process=True) as proxy:
        start_time = time.time()

        cb = EventCallback()

        event_id = proxy.subscribe_event("scalar_int", EventType.ATTR_CONF_EVENT, cb)

        def check_limits(attr_conf_event, min_value, max_value):
            attr_info = attr_conf_event.attr_conf
            assert attr_info.min_value == min_value
            assert attr_info.max_value == max_value

        time.sleep(A_BIT)
        events = cb.get_events()
        assert len(events) == 1
        ev = events[0]
        check_limits(ev, "Not specified", "Not specified")
        assert "scalar_int" in ev.attr_name
        assert ev.event == "attr_conf"
        assert not ev.err
        tv1 = ev.get_date()
        tv2 = ev.reception_date
        assert str(tv1) == str(tv2)
        assert start_time < tv1.totime() < start_time + 2 * A_BIT
        assert ev.errors == ()
        assert ev.device is proxy

        new_limits = [10, 15]
        proxy.change_min_max(new_limits)

        time.sleep(A_BIT)
        assert len(events) == 3
        check_limits(events[1], "10", "Not specified")
        check_limits(events[2], "10", "15")

        proxy.unsubscribe_event(event_id)


def test_attribute_configuration_event_set_errors():

    conf_event = AttrConfEventData()
    conf_event.errors = DevFailed(DevError())
    assert len(conf_event.errors) == 1



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
import os

from tango import ApiUtil


def test_get_env_var():

    key = "MYTEST"
    value = "abcd"

    os.environ[key] = value
    assert ApiUtil.get_env_var(key) == value



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
"""A simple TCP server for Tango attributes.

It runs on all interfaces on port 8888:

   $ python tango_tcp_server.py
   Serving on 0.0.0.0 port 8888

It can be accessed using netcat:

   $ ncat localhost 8888
   >>> sys/tg_test/1/ampli
   0.0
   >>> sys/tg_test/1/state
   RUNNING
   >>> sys/tg_test/1/nope
   DevFailed[
   DevError[
        desc = Attribute nope is not supported by device sys/tg_test/1
      origin = AttributeProxy::real_constructor()
      reason = API_UnsupportedAttribute
    severity = ERR]
    ]
   >>> ...
"""

import asyncio
from tango.asyncio import AttributeProxy


async def handle_echo(reader, writer):
    # Write the cursor
    writer.write(b">>> ")
    # Loop over client request
    async for line in reader:
        request = line.decode().strip()
        # Get attribute value using asyncio green mode
        try:
            proxy = await AttributeProxy(request)
            attr_value = await proxy.read()
            reply = str(attr_value.value)
        # Catch exception if something goes wrong
        except Exception as exc:
            reply = str(exc)
        # Reply to client
        writer.write(reply.encode() + b"\n" + b">>> ")
    # Close communication
    writer.close()


async def start_serving():
    server = await asyncio.start_server(handle_echo, "0.0.0.0", 8888)
    print("Serving on {} port {}".format(*server.sockets[0].getsockname()))
    return server


async def stop_serving(server):
    server.close()
    await server.wait_closed()


def main():
    # Start the server
    loop = asyncio.get_event_loop()
    server = loop.run_until_complete(start_serving())
    # Serve requests until Ctrl+C is pressed
    try:
        loop.run_forever()
    except KeyboardInterrupt:
        pass
    # Close the server
    loop.run_until_complete(stop_serving(server))
    loop.close()


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
import numpy as np

try:
    import numpy.typing as npt
except ImportError:
    npt = None

import pytest

from typing import Any
from tango import (
    DevFailed,
    PyTangoUserWarning,  # noqa
)
from tango.server import Device
from tango.server import command, device_property, attribute
from tango.test_utils import DeviceTestContext

from tango.test_utils import (
    assert_close,
    convert_dtype_to_typing_hint,
)


def test_device_property_no_default(
    general_typed_values,
):
    dtype, values, expected = general_typed_values
    patched_dtype = dtype if dtype != (bool,) else (int,)
    value = values[1]

    class TestDevice(Device):

        prop_without_db_value = device_property(dtype=dtype)
        prop_with_db_value = device_property(dtype=dtype)

        @command(dtype_out=bool)
        def is_prop_without_db_value_set_to_none(self):
            return self.prop_without_db_value is None

        @command(dtype_out=patched_dtype)
        def get_prop_with_db_value(self):
            return self.prop_with_db_value

    with DeviceTestContext(
        TestDevice, properties={"prop_with_db_value": value}
    ) as proxy:
        assert proxy.is_prop_without_db_value_set_to_none()
        assert_close(proxy.get_prop_with_db_value(), expected(value))


def test_device_property_with_typing(general_typed_values):
    dtype, values, expected = general_typed_values
    patched_dtype = dtype if dtype != (bool,) else (int,)
    value = values[1]

    tuple_hint, list_hint, _, _ = convert_dtype_to_typing_hint(dtype)

    class TestDevice(Device):
        prop_tuple_hint: tuple_hint = device_property()

        prop_list_hint: list_hint = device_property()

        prop_user_type_has_priority: dict = device_property(dtype=dtype)

        @command(dtype_out=patched_dtype)
        def get_prop_tuple_hint(self):
            return self.prop_tuple_hint

        @command(dtype_out=patched_dtype)
        def get_prop_list_hint(self):
            return self.prop_list_hint

        @command(dtype_out=patched_dtype)
        def get_prop_user_type_has_priority(self):
            return self.prop_user_type_has_priority

    with DeviceTestContext(
        TestDevice,
        properties={
            "prop_tuple_hint": value,
            "prop_list_hint": value,
            "prop_user_type_has_priority": value,
        },
    ) as proxy:
        assert_close(proxy.get_prop_tuple_hint(), expected(value))
        assert_close(proxy.get_prop_list_hint(), expected(value))
        assert_close(proxy.get_prop_user_type_has_priority(), expected(value))


if npt:

    def test_device_property_with_numpy_typing(command_numpy_typed_values):
        type_hint, dformat, value, expected = command_numpy_typed_values
        if type_hint in [np.uint8, npt.NDArray[np.uint8]]:
            pytest.xfail("Does not work for some reason")

        class TestDevice(Device):
            prop: type_hint = device_property()

            @command(dformat_out=dformat)
            def get_prop(self) -> type_hint:
                return self.prop

        with DeviceTestContext(TestDevice, properties={"prop": value}) as proxy:
            assert_close(proxy.get_prop(), expected(value))


@pytest.mark.parametrize("input_type", [str, Any])
def test_device_property_with_default_value(general_typed_values, input_type):
    dtype, values, expected = general_typed_values
    patched_dtype = dtype if dtype != (bool,) else (int,)

    if isinstance(input_type, str) and isinstance(values[0], list):
        default_set = [str(v) for v in values[0]]
    elif isinstance(input_type, str):
        default_set = str(values[0])
    else:
        default_set = values[0]
    default_expected = values[0]
    value = values[1]

    class TestDevice(Device):
        prop_without_db_value = device_property(dtype=dtype, default_value=default_set)
        prop_with_db_value = device_property(dtype=dtype, default_value=default_set)

        @command(dtype_out=patched_dtype)
        def get_prop_without_db_value(self):
            return self.prop_without_db_value

        @command(dtype_out=patched_dtype)
        def get_prop_with_db_value(self):
            return self.prop_with_db_value

    with DeviceTestContext(
        TestDevice, properties={"prop_with_db_value": value}
    ) as proxy:
        assert_close(proxy.get_prop_without_db_value(), expected(default_expected))
        assert_close(proxy.get_prop_with_db_value(), expected(value))


def test_device_get_device_properties_when_init_device():
    class TestDevice(Device):
        _got_properties = False

        def get_device_properties(self, *args, **kwargs):
            super().get_device_properties(*args, **kwargs)
            self._got_properties = True

        @attribute(dtype=bool)
        def got_properties(self):
            return self._got_properties

    with DeviceTestContext(TestDevice) as proxy:
        assert proxy.got_properties


@pytest.mark.parametrize("set_default", [True, False])
def test_mandatory_device_property_with_db_value_succeeds(set_default):

    class TestDevice(Device):

        prop = device_property(dtype=int, mandatory=True)

        @command(dtype_out=int)
        def get_prop(self):
            return self.prop

    if set_default:
        with DeviceTestContext(TestDevice, properties={"prop": 1}) as proxy:
            assert proxy.get_prop() == 1
    else:
        with pytest.raises(DevFailed) as context:
            with DeviceTestContext(TestDevice):
                pass
        assert "Device property prop is mandatory" in str(context.value)



import time

import tango


class EventManager:
    def __init__(self, dp):
        self._deviceProxy = dp
        if dp is not None:
            print("Subscribed to Interface Change Events")
            self._event_id = dp.subscribe_event(
                tango.EventType.INTERFACE_CHANGE_EVENT, self
            )

    def unsubscribe(self):
        self._deviceProxy.unsubscribe_event(self._event_id)

    def push_event(self, ev):
        print("Event -----push_event-----------------")
        print("Timestamp:      ", ev.reception_date)
        print("Event type:     ", ev.event)
        print("Device server:  ", ev.device)
        print("Event error:    ", ev.err)
        if ev.err:
            print("Caught pipe exception")
            err = ev.errors[0]
            print("Error desc:     ", err.desc)
            print("Error origin:   ", err.origin)
            print("Error reason:   ", err.reason)
            print("Error severity: ", err.severity)
        else:
            if ev.cmd_list is not None:
                print("Number of commands   ", len(ev.cmd_list))
                for i in range(len(ev.cmd_list)):
                    cmdInfo = ev.cmd_list[i]
                    print("cmd  -----> ", cmdInfo.cmd_name)
            else:
                print("Number of commands    0")

            if ev.att_list is not None:
                print("Number of attributes ", len(ev.att_list))
                for i in range(len(ev.att_list)):
                    attInfo = ev.att_list[i]
                    print("att  -----> ", attInfo.name)
            else:
                print("Number of attributes  0")

            print("Device started ", ev.dev_started)


def main():
    dev = tango.DeviceProxy("ifchangeServer/tango/1")
    EventManager(dev)
    time.sleep(3000.0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
#
# client example: writing an attribute, also with simplified syntax

import sys
from tango import DeviceProxy, DevFailed, Except

dev_name = "sys/tg_test/1"
attr_name = "double_scalar"

# step 1 create the device proxy
try:
    dev = DeviceProxy(dev_name)
    print(f"proxy for {dev_name} created")
except DevFailed as ex:
    Except.print_exception(ex)
    sys.exit(1)

# step 2 write the attribute
try:
    dev.write_attribute(attr_name, 5.678)  # write the attribute directly
    print(f"write_attribute {attr_name!r} OK")
except DevFailed as ex:
    Except.print_exception(ex)
    print("failed to write attribute")
    sys.exit(1)

# use simplified syntax
try:
    dev.attr_name = 0.618  # WRONG! you are defining new attribute attr_name
    # in object dev with value 0.618
    dev.double_scalar = 0.619
    print(f"write_attribute SIMPLIFIED {attr_name!r} OK")
except DevFailed as ex:
    Except.print_exception(ex)
    print("failed to write attribute with simplified syntax")
    sys.exit(1)

"""
Typical output:

➜  training git:(develop) ✗ docker-compose exec cli /training/client/writing01.py
proxy for sys/tg_test/1 created
write_attribute 'double_scalar' OK
write_attribute SIMPLIFIED 'double_scalar' OK
"""



import os
from pybind11_stubgen import main

for p in os.environ.get("PATH", "").split(";"):
    try:
        os.add_dll_directory(p)
    except:  # noqa: E722
        print(f"Failed adding this dll directory: {p}")
        print("Continuing...")

main()



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

"""
This is an internal PyTango module.
"""


import atexit
import collections.abc
import contextlib
import logging
import os
import re
import socket
import sys
import types
import numbers
import inspect
import enum
import threading
import warnings
import functools
import weakref
import time

import typing
import numpy as np

from argparse import HelpFormatter
from collections import namedtuple
from contextvars import ContextVar
from urllib.parse import urlparse, urlunparse

from packaging.version import Version

from tango import (
    __auto_die_callbacks_holder,
    AttrQuality,
    StdStringVector,
    StdDoubleVector,
    DbData,
    DbDatum,
    DbDevInfos,
    DbDevExportInfos,
    CmdArgType,
    AttrDataFormat,
    EventData,
    EventReason,
    AttrConfEventData,
    DataReadyEventData,
    DevFailed,
    DevState,
    DevIntrChangeEventData,
    Database,
    ApiUtil,
    EnsureOmniThread,
    DeviceData,
    DeviceDataList,
    DeviceProxy,
    Except,
    ErrSeverity,
    LockerLanguage,
    is_omni_thread,
)

from tango import _tango

from tango._tango import _telemetry, __CallBackAutoDie, __Group
from tango.constants import (
    AlrmValueNotSpec,
    StatusNotSet,
    TgLibVers,
    TELEMETRY_SUPPORTED,
)
from tango.release import Release

__all__ = (
    "requires_pytango",
    "requires_tango",
    "is_pure_str",
    "is_seq",
    "is_non_str_seq",
    "is_integer",
    "is_number",
    "is_scalar_type",
    "is_array_type",
    "is_numerical_type",
    "is_int_type",
    "is_float_type",
    "is_bool_type",
    "is_binary_type",
    "is_str_type",
    "obj_2_str",
    "seqStr_2_obj",
    "scalar_to_array_type",
    "document_method",
    "document_static_method",
    "CaselessList",
    "CaselessDict",
    "EventCallback",
    "AsyncEventCallback",
    "get_home",
    "from_version_str_to_hex_str",
    "from_version_str_to_int",
    "seq_2_StdStringVector",
    "StdStringVector_2_seq",
    "dir2",
    "TO_TANGO_TYPE",
    "ensure_binary",
    "_get_device_fqtrl_if_necessary",
    "_get_test_context_tango_host_fqtrl",
    "_set_test_context_tango_host_fqtrl",
    "_clear_test_context_tango_host_fqtrl",
    "InvalidTangoHostTrlError",
    "PyTangoUserWarning",
    "parse_type_hint",
    "PyTangoThreadPoolExecutor",
    "set_telemetry_tracer_provider_factory",
    "get_telemetry_tracer_provider_factory",
    "_create_device_telemetry_tracer",
    "_get_non_tango_source_location",
    "_get_current_otel_context",
    "_span_to_cpptango",
    "_telemetry_active",
    "_trace_client",
    "_DummyTracer",
    "_get_command_inout_param",
    "parameter_2_dbdata",
)

__docformat__ = "restructuredtext"

# Types

__str_klasses = (str,)
__int_klasses = (int, np.integer)
__number_klasses = (numbers.Number, np.number)
__seq_klasses = (collections.abc.Sequence, bytearray, StdStringVector, np.ndarray)

_scalar_int_types = (
    CmdArgType.DevShort,
    CmdArgType.DevUShort,
    CmdArgType.DevLong,
    CmdArgType.DevULong,
    CmdArgType.DevLong64,
    CmdArgType.DevULong64,
)

_scalar_float_types = (
    CmdArgType.DevFloat,
    CmdArgType.DevDouble,
)

_scalar_numerical_types = _scalar_int_types + _scalar_float_types

_scalar_str_types = (
    CmdArgType.DevString,
    CmdArgType.ConstDevString,
)

_scalar_bool_types = (CmdArgType.DevBoolean,)

_scalar_types = (
    _scalar_numerical_types
    + _scalar_str_types
    + _scalar_bool_types
    + (
        CmdArgType.DevEncoded,
        CmdArgType.DevUChar,
    )
)

_array_int_types = (
    CmdArgType.DevVarShortArray,
    CmdArgType.DevVarUShortArray,
    CmdArgType.DevVarLongArray,
    CmdArgType.DevVarULongArray,
    CmdArgType.DevVarLong64Array,
    CmdArgType.DevVarULong64Array,
)

_array_float_types = (CmdArgType.DevVarFloatArray, CmdArgType.DevVarDoubleArray)

_array_numerical_types = _array_int_types + _array_float_types

_array_str_types = (CmdArgType.DevVarStringArray,)

_array_bool_types = (CmdArgType.DevVarBooleanArray,)

_array_types = (
    _array_numerical_types
    + _array_bool_types
    + _array_str_types
    + (
        CmdArgType.DevVarCharArray,
        CmdArgType.DevVarDoubleStringArray,
        CmdArgType.DevVarLongStringArray,
    )
)

_binary_types = (
    CmdArgType.DevEncoded,
    CmdArgType.DevVarCharArray,
)


FROM_TANGO_TO_NUMPY_TYPE = {
    CmdArgType.DevBoolean: np.bool_,
    CmdArgType.DevUChar: np.ubyte,
    CmdArgType.DevShort: np.short,
    CmdArgType.DevUShort: np.ushort,
    CmdArgType.DevLong: np.int32,
    CmdArgType.DevULong: np.uint32,
    CmdArgType.DevLong64: np.int64,
    CmdArgType.DevULong64: np.uint64,
    CmdArgType.DevString: str,
    CmdArgType.DevDouble: np.float64,
    CmdArgType.DevFloat: np.float32,
}


def __build_to_tango_type():
    ret = {
        int: CmdArgType.DevLong64,
        str: CmdArgType.DevString,
        bool: CmdArgType.DevBoolean,
        bytearray: CmdArgType.DevEncoded,
        float: CmdArgType.DevDouble,
        chr: CmdArgType.DevUChar,
        None: CmdArgType.DevVoid,
        "int": CmdArgType.DevLong64,
        "int16": CmdArgType.DevShort,
        "int32": CmdArgType.DevLong,
        "int64": CmdArgType.DevLong64,
        "uint": CmdArgType.DevULong64,
        "uint16": CmdArgType.DevUShort,
        "uint32": CmdArgType.DevULong,
        "uint64": CmdArgType.DevULong64,
        "str": CmdArgType.DevString,
        "string": CmdArgType.DevString,
        "text": CmdArgType.DevString,
        "bool": CmdArgType.DevBoolean,
        "boolean": CmdArgType.DevBoolean,
        "bytes": CmdArgType.DevEncoded,
        "bytearray": CmdArgType.DevEncoded,
        "float": CmdArgType.DevDouble,
        "float32": CmdArgType.DevFloat,
        "float64": CmdArgType.DevDouble,
        "double": CmdArgType.DevDouble,
        "byte": CmdArgType.DevUChar,
        "chr": CmdArgType.DevUChar,
        "char": CmdArgType.DevUChar,
        "None": CmdArgType.DevVoid,
        "state": CmdArgType.DevState,
        "enum": CmdArgType.DevEnum,
    }

    for key in dir(CmdArgType):
        if key.startswith("Dev"):
            value = getattr(CmdArgType, key)
            ret[key] = ret[value] = value

        for key, value in FROM_TANGO_TO_NUMPY_TYPE.items():
            ret[value] = key
    return ret


TO_TANGO_TYPE = __build_to_tango_type()

_scalar_to_array_type = {
    CmdArgType.DevBoolean: CmdArgType.DevVarBooleanArray,
    CmdArgType.DevUChar: CmdArgType.DevVarCharArray,
    CmdArgType.DevShort: CmdArgType.DevVarShortArray,
    CmdArgType.DevUShort: CmdArgType.DevVarUShortArray,
    CmdArgType.DevLong: CmdArgType.DevVarLongArray,
    CmdArgType.DevULong: CmdArgType.DevVarULongArray,
    CmdArgType.DevLong64: CmdArgType.DevVarLong64Array,
    CmdArgType.DevULong64: CmdArgType.DevVarULong64Array,
    CmdArgType.DevFloat: CmdArgType.DevVarFloatArray,
    CmdArgType.DevDouble: CmdArgType.DevVarDoubleArray,
    CmdArgType.DevString: CmdArgType.DevVarStringArray,
    CmdArgType.ConstDevString: CmdArgType.DevVarStringArray,
}

# add derived scalar types to scalar to array map
for k, v in TO_TANGO_TYPE.items():
    if v in _scalar_to_array_type:
        _scalar_to_array_type[k] = _scalar_to_array_type[v]

__NO_STR_VALUE = AlrmValueNotSpec, StatusNotSet

__device_classes = None

bool_ = lambda value_str: value_str.lower() == "true"


def __import(name):
    __import__(name)
    return sys.modules[name]


def __requires(package_name, min_version=None, conflicts=(), software_name="Software"):
    package_name_l = package_name.lower()
    if package_name_l == "pytango":
        curr_version = Version(Release.version)
    elif package_name_l == "tango":
        curr_version = Version(TgLibVers)
    else:
        try:
            package = __import(package_name)
            curr_version = Version(package.__version__)
        except ImportError:
            msg = f"Could not find package {package_name} required by {software_name}"
            raise Exception(msg)
        except Exception:
            msg = f"Error importing package {package_name} required by {software_name}"
            raise Exception(msg)

    if min_version is not None:
        min_version = Version(min_version)
        if min_version > curr_version:
            msg = f"{software_name} requires {package_name} {min_version} but {curr_version} installed"
            raise Exception(msg)

    conflicts = map(Version, conflicts)
    if curr_version in conflicts:
        msg = f"{software_name} cannot run with {package_name} {curr_version}"
        raise Exception(msg)
    return True


def requires_pytango(min_version=None, conflicts=(), software_name="Software"):
    """
    Determines if the required PyTango version for the running
    software is present. If not an exception is thrown.
    Example usage::

        from tango import requires_pytango

        requires_pytango('7.1', conflicts=['8.1.1'], software_name='MyDS')

    :param min_version:
        minimum PyTango version [default: None, meaning no minimum
        required]. If a string is given, it must be in the valid
        version number format
        (see: :class:`~packaging.version.Version`)
    :type min_version:
        None, str, :class:`~packaging.version.Version`
    :param conflicts:
        a sequence of PyTango versions which conflict with the
        software using it
    :type conflicts:
        seq<str|Version>
    :param software_name:
        software name using tango. Used in the exception message
    :type software_name: str

    :raises Exception: if the required PyTango version is not met

    New in PyTango 8.1.4
    """
    return __requires(
        "pytango",
        min_version=min_version,
        conflicts=conflicts,
        software_name=software_name,
    )


def requires_tango(min_version=None, conflicts=(), software_name="Software"):
    """
    Determines if the required cppTango version for the running
    software is present. If not an exception is thrown.
    Example usage::

        from tango import requires_tango

        requires_tango('7.1', conflicts=['8.1.1'], software_name='MyDS')

    :param min_version:
        minimum Tango version [default: None, meaning no minimum
        required]. If a string is given, it must be in the valid
        version number format
        (see: :class:`~packaging.version.Version`)
    :type min_version:
        None, str, :class:`~packaging.version.Version`
    :param conflicts:
        a sequence of Tango versions which conflict with the
        software using it
    :type conflicts:
        seq<str|Version>
    :param software_name:
        software name using Tango. Used in the exception message
    :type software_name: str

    :raises Exception: if the required Tango version is not met

    New in PyTango 8.1.4
    """
    return __requires(
        "Tango",
        min_version=min_version,
        conflicts=conflicts,
        software_name=software_name,
    )


def get_tango_device_classes():
    global __device_classes
    if __device_classes is None:
        __device_classes = [_tango.DeviceImpl]
        i = 2
        while True:
            dc = f"Device_{i}Impl"
            try:
                __device_classes.append(getattr(_tango, dc))
                i = i + 1
            except AttributeError:
                break
    return __device_classes


def get_latest_device_class():
    return get_tango_device_classes()[-1]


def set_complex_value(attr, value):
    if not isinstance(value, tuple):
        attr.set_value(value)
    else:
        n_vals = len(value)
        if (n_vals > 2 and isinstance(value[2], AttrQuality)) or (
            n_vals > 3
            and attr.get_data_type() == CmdArgType.DevEncoded
            and isinstance(value[3], AttrQuality)
        ):
            attr.set_value_date_quality(*value)
        else:
            attr.set_value(value)


def __get_tango_type(obj):
    if is_non_str_seq(obj):
        tg_type, tg_format = get_tango_type(obj[0])
        tg_format = AttrDataFormat(int(tg_format) + 1)
        return tg_type, tg_format
    elif is_pure_str(obj):
        r = CmdArgType.DevString
    elif isinstance(obj, DevState):
        r = CmdArgType.DevState
    elif isinstance(obj, bool):
        r = CmdArgType.DevBoolean
    elif isinstance(obj, __int_klasses):
        r = CmdArgType.DevLong64
    elif isinstance(obj, __number_klasses):
        r = CmdArgType.DevDouble
    else:
        raise TypeError(f"Unsupported type {type(obj)}")
    return r, AttrDataFormat.SCALAR


def get_tango_type(obj):
    try:
        ndim, dtype = obj.ndim, str(obj.dtype)
        if ndim > 2:
            raise TypeError(
                f"cannot translate numpy array with {obj.ndim} "
                f"dimensions to tango type"
            )
        return TO_TANGO_TYPE[dtype], AttrDataFormat(ndim)
    except AttributeError:
        return __get_tango_type(obj)


def get_attribute_type_format(dtype, dformat, enum_labels):
    if is_devstate(dtype) or is_devstate_seq(dtype):
        _dtype = dtype
        dtype = CmdArgType.DevState

        while is_devstate_seq(_dtype):
            _dtype = _dtype[0]
            dtype = (dtype,)
    elif is_enum(dtype) or is_enum_seq(dtype):
        if enum_labels:
            raise TypeError(
                "For dtype of enum.Enum, (enum.Enum,) or ((enum.Enum,),) the enum_labels must not "
                f"be specified - dtype: {dtype}, enum_labels: {enum_labels}."
            )
        _dtype = dtype
        dtype = CmdArgType.DevEnum

        while is_enum_seq(_dtype):
            _dtype = _dtype[0]
            dtype = (dtype,)

        enum_labels = get_enum_labels(_dtype)

    return *get_tango_type_format(dtype, dformat, "attribute"), enum_labels


def get_tango_type_format(dtype, dformat=None, caller=None):
    if dformat is None:
        dformat = AttrDataFormat.SCALAR
        if is_non_str_seq(dtype):
            if len(dtype):
                dtype = dtype[0]
                dformat = AttrDataFormat.SPECTRUM
                if is_non_str_seq(dtype):
                    if len(dtype):
                        dtype = dtype[0]
                        dformat = AttrDataFormat.IMAGE
                    elif caller == "attribute":
                        raise TypeError(
                            "Image attribute type must be specified as ((<dtype>,),)"
                        )
            elif caller == "attribute":
                raise TypeError(
                    "Spectrum attribute type must be specified as (<dtype>,)"
                )

    try:
        tango_type = TO_TANGO_TYPE[dtype]
    except KeyError:
        if isinstance(dtype, str):
            raise RuntimeError(
                f"Cannot translate {dtype!r} to TANGO type. See documentation for the allowed types. "
                'If you are declaring type with type hints and using "from __future__ import annotations", '
                "please read documentation to learn about limitations"
            )
        else:
            raise RuntimeError(
                f"Cannot translate {dtype!r} to TANGO type. See documentation for the allowed types"
            )

    return tango_type, dformat


def __has_ellipsis_in_types(dtype):
    try:
        list(typing.get_args(dtype)).remove(Ellipsis)
        return True
    except ValueError:
        pass

    return False


def __check_types_uniformity(dtype):
    all_types = list(set(typing.get_args(dtype)))
    if Ellipsis in all_types:
        all_types.remove(Ellipsis)
    if len(all_types) > 1:
        return False
    for sub_type in typing.get_args(dtype):
        if typing.get_origin(sub_type) in [list, tuple]:
            if not __check_types_uniformity(sub_type):
                return False
    return True


def eval_in_fallback_frames(dtype_str, try_depth):
    """
    Tries to eval a string at `try_depth` frames back from the caller.
    """

    def get_frame_at_depth(depth):
        # We go back depth + 2 times:
        #  +1 to get to eval_in_fallback_frames's caller
        #  +1 to get to this helper's caller (which is eval_in_fallback_frames)
        #  +depth to get to the target frame.
        # A simpler way: start at f_back.f_back and loop 'depth' times.

        current_frame = inspect.currentframe().f_back.f_back
        for _ in range(depth):
            if current_frame:
                current_frame = current_frame.f_back
            else:
                # Stack is not deep enough
                return None
        return current_frame

    frame_to_try = None
    try:
        frame_to_try = get_frame_at_depth(try_depth)
        if frame_to_try is None:
            raise IndexError()
        return eval(dtype_str, frame_to_try.f_globals, frame_to_try.f_locals)
    finally:
        # Ensure all created frame references are deleted
        if frame_to_try:
            del frame_to_try


def parse_type_hint(annotation, caller):
    dtype = annotation
    if isinstance(dtype, str):
        depth = 1
        while depth <= 6:  # user code should be before 6th frame
            try:
                dtype = eval_in_fallback_frames(dtype, depth)
                break
            except NameError:
                depth += 1
            except IndexError:
                break

    dformat = AttrDataFormat.SCALAR
    max_x, max_y = None, None
    if typing.get_origin(dtype) in [list, tuple]:
        n_elements = len(typing.get_args(dtype))
        if (
            n_elements in [2, 4]
            and typing.get_args(dtype)[0] is str
            and typing.get_args(dtype)[1] in [bytes, bytearray]
        ):
            return "DevEncoded", dformat, max_x, max_y
        if (
            n_elements in [2, 4]
            and typing.get_args(dtype)[0] in [tuple[int], list[int]]
            and typing.get_args(dtype)[1] in [tuple[str], list[str]]
        ):
            return "DevVarLongStringArray", dformat, max_x, max_y
        if (
            n_elements in [2, 4]
            and typing.get_args(dtype)[0] in [tuple[float], list[float]]
            and typing.get_args(dtype)[1] in [tuple[str], list[str]]
        ):
            return "DevVarDoubleStringArray", dformat, max_x, max_y
        if (
            n_elements == 3
            and typing.get_args(dtype)[1] is float
            and typing.get_args(dtype)[2] == AttrQuality
        ):
            dtype = typing.get_args(dtype)[0]
    if typing.get_origin(dtype) == np.ndarray:
        dtype = typing.get_args(typing.get_args(dtype)[1])[0]
        dformat = None
        if caller in ["property", "command"]:
            dtype = (dtype,)
    if typing.get_origin(dtype) in [list, tuple]:
        if not __has_ellipsis_in_types(dtype):
            max_x = len(typing.get_args(dtype))
        types_are_uniform = __check_types_uniformity(dtype)
        dtype = typing.get_args(dtype)[0]
        dformat = (
            AttrDataFormat.IMAGE
            if typing.get_origin(dtype) in [list, tuple]
            else AttrDataFormat.SPECTRUM
        )

        if caller in ["property", "command"] and dformat == AttrDataFormat.IMAGE:
            raise RuntimeError(f"{caller.capitalize()} does not support IMAGE type")

        if not types_are_uniform:
            if caller in ["property", "command"]:
                raise RuntimeError(
                    f"PyTango does not support mixed types in SPECTRUM {caller}"
                )
            else:
                if dformat == AttrDataFormat.IMAGE:
                    raise RuntimeError(
                        "PyTango does not support mixed types in IMAGE attributes"
                    )
                else:
                    raise RuntimeError(
                        "PyTango does not support mixed types in SPECTRUM attributes"
                    )

        if caller == "attribute":
            if dformat == AttrDataFormat.IMAGE:
                max_y = max_x
                if not __has_ellipsis_in_types(dtype):
                    max_x = len(typing.get_args(dtype))
                else:
                    max_x = None
                dtype = typing.get_args(dtype)[0]
        if caller == "property":
            dtype = (dtype,)

    return dtype, dformat, max_x, max_y


class EnumTypeError(Exception):
    """Invalid Enum class for use with DEV_ENUM."""


def get_enum_labels(enum_cls):
    """
    Return list of enumeration labels from Enum class.

    The list is useful when creating an attribute, for the
    `enum_labels` parameter.  The enumeration values are checked
    to ensure they are unique, start at zero, and increment by one.

    :param enum_cls: the Enum class to be inspected
    :type enum_cls: :py:obj:`enum.Enum`

    :return: List of label strings
    :rtype: :py:obj:`list`

    :raises EnumTypeError: in case the given class is invalid
    """
    if not issubclass(enum_cls, enum.Enum):
        raise EnumTypeError(f"Input class '{enum_cls}' must be derived from enum.Enum")

    # Check there are no duplicate labels
    try:
        enum.unique(enum_cls)
    except ValueError as exc:
        raise EnumTypeError(f"Input class '{enum_cls}' must be unique - {exc}")

    # Check the values start at 0, and increment by 1, since that is
    # assumed by tango's DEV_ENUM implementation.
    values = [member.value for member in enum_cls]
    if not values:
        raise EnumTypeError(f"Input class '{enum_cls}' has no members!")
    expected_value = 0
    for value in values:
        if value != expected_value:
            raise EnumTypeError(
                f"Enum values for '{enum_cls}' must start at 0 and "
                f"increment by 1.  Values: {values}"
            )
        expected_value += 1

    return [member.name for member in enum_cls]


def is_pure_str(obj):
    """
    Tells if the given object is a python string.

    In python 2.x this means any subclass of basestring.
    In python 3.x this means any subclass of str.

    :param obj: the object to be inspected
    :type obj: :py:obj:`object`

    :return: True is the given obj is a string or False otherwise
    :rtype: :py:obj:`bool`
    """
    return isinstance(obj, __str_klasses)


def is_seq(obj):
    """
    Tells if the given object is a python sequence.

    It will return True for any collections.Sequence (list, tuple,
    str, bytes, unicode), bytearray and (if numpy is enabled)
    numpy.ndarray

    :param obj: the object to be inspected
    :type obj: :py:obj:`object`

    :return: True is the given obj is a sequence or False otherwise
    :rtype: :py:obj:`bool`
    """
    return isinstance(obj, __seq_klasses)


def is_non_str_seq(obj):
    """
    Tells if the given object is a python sequence (excluding string
    sequences).

    It will return True for any collections.Sequence (list, tuple (and
    bytes in python3)), bytearray and (if numpy is enabled)
    numpy.ndarray

    :param obj: the object to be inspected
    :type obj: :py:obj:`object`

    :return: True is the given obj is a sequence or False otherwise
    :rtype: :py:obj:`bool`
    """
    return is_seq(obj) and not is_pure_str(obj)


def is_devstate(obj):
    return inspect.isclass(obj) and issubclass(obj, DevState)


def is_devstate_seq(obj):
    if is_non_str_seq(obj):
        while is_non_str_seq(obj):
            obj = obj[0]
        return is_devstate(obj)
    return False


def is_enum(obj):
    return inspect.isclass(obj) and issubclass(obj, enum.Enum)


def is_enum_seq(obj):
    if is_non_str_seq(obj):
        while is_non_str_seq(obj) and len(obj):
            obj = obj[0]
        return is_enum(obj)
    return False


def is_integer(obj):
    """
    Tells if the given object is a python integer.

    It will return True for any int, long (in python 2) and
    (if numpy is enabled) numpy.integer

    :param obj: the object to be inspected
    :type obj: :py:obj:`object`

    :return:
        True is the given obj is a python integer or False otherwise
    :rtype: :py:obj:`bool`
    """
    return isinstance(obj, __int_klasses) and not isinstance(obj, bool)


def is_number(obj):
    """
    Tells if the given object is a python number.

    It will return True for any numbers.Number and (if numpy is
    enabled) numpy.number

    :param obj: the object to be inspected
    :type obj: :py:obj:`object`

    :return:
        True is the given obj is a python number or False otherwise
    :rtype: :py:obj:`bool`
    """
    return isinstance(obj, __number_klasses)


def is_scalar(tg_type):
    """Tells if the given tango type is a scalar

    :param tg_type: tango type
    :type tg_type: :class:`tango.CmdArgType`

    :return: True if the given tango type is a scalar or False otherwise
    :rtype: :py:obj:`bool`
    """

    global _scalar_types
    return tg_type in _scalar_types


is_scalar_type = is_scalar


def is_array(tg_type):
    """Tells if the given tango type is an array type

    :param tg_type: tango type
    :type tg_type: :class:`tango.CmdArgType`

    :return: True if the given tango type is an array type or False otherwise
    :rtype: :py:obj:`bool`
    """
    global _array_types
    return tg_type in _array_types


is_array_type = is_array


def is_numerical(tg_type, inc_array=False):
    """Tells if the given tango type is numerical

    :param tg_type: tango type
    :type tg_type: :class:`tango.CmdArgType`
    :param inc_array: (optional, default is False) determines if include array
                      in the list of checked types
    :type inc_array: :py:obj:`bool`

    :return: True if the given tango type is a numerical or False otherwise
    :rtype: :py:obj:`bool`
    """
    global _scalar_numerical_types, _array_numerical_types
    if tg_type in _scalar_numerical_types:
        return True
    if not inc_array:
        return False
    return tg_type in _array_numerical_types


is_numerical_type = is_numerical


def is_int(tg_type, inc_array=False):
    """Tells if the given tango type is integer

    :param tg_type: tango type
    :type tg_type: :class:`tango.CmdArgType`
    :param inc_array: (optional, default is False) determines if include array
                      in the list of checked types
    :type inc_array: :py:obj:`bool`

    :return: True if the given tango type is integer or False otherwise
    :rtype: :py:obj:`bool`
    """
    global _scalar_int_types, _array_int_types
    if tg_type in _scalar_int_types:
        return True
    if not inc_array:
        return False
    return tg_type in _array_int_types


is_int_type = is_int


def is_float(tg_type, inc_array=False):
    """Tells if the given tango type is float

    :param tg_type: tango type
    :type tg_type: :class:`tango.CmdArgType`
    :param inc_array: (optional, default is False) determines if include array
                      in the list of checked types
    :type inc_array: :py:obj:`bool`

    :return: True if the given tango type is float or False otherwise
    :rtype: :py:obj:`bool`
    """
    global _scalar_float_types, _array_float_types
    if tg_type in _scalar_float_types:
        return True
    if not inc_array:
        return False
    return tg_type in _array_float_types


is_float_type = is_float


def is_bool(tg_type, inc_array=False):
    """Tells if the given tango type is boolean

    :param tg_type: tango type
    :type tg_type: :class:`tango.CmdArgType`
    :param inc_array: (optional, default is False) determines if include array
                      in the list of checked types
    :type inc_array: :py:obj:`bool`

    :return: True if the given tango type is boolean or False otherwise
    :rtype: :py:obj:`bool`
    """
    global _scalar_bool_types, _array_bool_types
    if tg_type in _scalar_bool_types:
        return True
    if not inc_array:
        return False
    return tg_type in _array_bool_types


is_bool_type = is_bool


def is_str(tg_type, inc_array=False):
    """Tells if the given tango type is string

    :param tg_type: tango type
    :type tg_type: :class:`tango.CmdArgType`
    :param inc_array: (optional, default is False) determines if include array
                      in the list of checked types
    :type inc_array: :py:obj:`bool`

    :return: True if the given tango type is string or False otherwise
    :rtype: :py:obj:`bool`
    """
    global _scalar_str_types, _array_str_types
    if tg_type in _scalar_str_types:
        return True
    if not inc_array:
        return False
    return tg_type in _array_str_types


is_str_type = is_str


def is_binary(tg_type, inc_array=False):
    """Tells if the given tango type is binary

    :param tg_type: tango type
    :type tg_type: :class:`tango.CmdArgType`
    :param inc_array: (optional, default is False) determines if include array
                      in the list of checked types
    :type inc_array: :py:obj:`bool`

    :return: True if the given tango type is binary or False otherwise
    :rtype: :py:obj:`bool`
    """
    global _binary_types
    return tg_type in _binary_types


is_binary_type = is_binary


def seq_2_StdStringVector(seq, vec=None):
    """Converts a python sequence<str> object to a :class:`tango.StdStringVector`

    :param seq: the sequence of strings
    :type seq: sequence<:py:obj:`str`>
    :param vec: (optional, default is None) an :class:`tango.StdStringVector`
                to be filled. If None is given, a new :class:`tango.StdStringVector`
                is created
    :return: a :class:`tango.StdStringVector` filled with the same contents as seq
    :rtype: :class:`tango.StdStringVector`
    """
    if vec is None:
        if isinstance(seq, StdStringVector):
            return seq
        vec = StdStringVector()
    if not isinstance(vec, StdStringVector):
        raise TypeError("vec must be a tango.StdStringVector")
    for e in seq:
        vec.append(str(e))
    return vec


def StdStringVector_2_seq(vec, seq=None):
    """Converts a :class:`tango.StdStringVector` to a python sequence<str>

    :param seq: the :class:`tango.StdStringVector`
    :type seq: :class:`tango.StdStringVector`
    :param vec: (optional, default is None) a python sequence to be filled.
                 If None is given, a new list is created
    :return: a python sequence filled with the same contents as seq
    :rtype: sequence<str>
    """
    if seq is None:
        seq = []
    if not isinstance(vec, StdStringVector):
        raise TypeError("vec must be a tango.StdStringVector")
    for e in vec:
        seq.append(str(e))
    return seq


def seq_2_StdDoubleVector(seq, vec=None):
    """Converts a python sequence<float> object to a :class:`tango.StdDoubleVector`

    :param seq: the sequence of floats
    :type seq: sequence<:py:obj:`float`>
    :param vec: (optional, default is None) an :class:`tango.StdDoubleVector`
                to be filled. If None is given, a new :class:`tango.StdDoubleVector`
                is created
    :return: a :class:`tango.StdDoubleVector` filled with the same contents as seq
    :rtype: :class:`tango.StdDoubleVector`
    """
    if vec is None:
        if isinstance(seq, StdDoubleVector):
            return seq
        vec = StdDoubleVector()
    if not isinstance(vec, StdDoubleVector):
        raise TypeError("vec must be a tango.StdDoubleVector")
    for e in seq:
        vec.append(float(e))
    return vec


def StdDoubleVector_2_seq(vec, seq=None):
    """Converts a :class:`tango.StdDoubleVector` to a python sequence<float>

    :param seq: the :class:`tango.StdDoubleVector`
    :type seq: :class:`tango.StdDoubleVector`
    :param vec: (optional, default is None) a python sequence to be filled.
                 If None is given, a new list is created
    :return: a python sequence filled with the same contents as seq
    :rtype: sequence<float>
    """
    if seq is None:
        seq = []
    if not isinstance(vec, StdDoubleVector):
        raise TypeError("vec must be a tango.StdDoubleVector")
    for e in vec:
        seq.append(float(e))
    return seq


def seq_2_DbDevInfos(seq, vec=None):
    """Converts a python sequence<DbDevInfo> object to a :class:`tango.DbDevInfos`

    :param seq: the sequence of DbDevInfo
    :type seq: sequence<DbDevInfo>
    :param vec: (optional, default is None) an :class:`tango.DbDevInfos`
                to be filled. If None is given, a new :class:`tango.DbDevInfos`
                is created
    :return: a :class:`tango.DbDevInfos` filled with the same contents as seq
    :rtype: :class:`tango.DbDevInfos`
    """
    if vec is None:
        if isinstance(seq, DbDevInfos):
            return seq
        vec = DbDevInfos()
    if not isinstance(vec, DbDevInfos):
        raise TypeError("vec must be a tango.DbDevInfos")
    for e in seq:
        vec.append(e)
    return vec


def seq_2_DbDevExportInfos(seq, vec=None):
    """Converts a python sequence<DbDevExportInfo> object to a :class:`tango.DbDevExportInfos`

    :param seq: the sequence of DbDevExportInfo
    :type seq: sequence<DbDevExportInfo>
    :param vec: (optional, default is None) an :class:`tango.DbDevExportInfos`
                to be filled. If None is given, a new :class:`tango.DbDevExportInfos`
                is created
    :return: a :class:`tango.DbDevExportInfos` filled with the same contents as seq
    :rtype: :class:`tango.DbDevExportInfos`
    """
    if vec is None:
        if isinstance(seq, DbDevExportInfos):
            return seq
        vec = DbDevExportInfos()
    if not isinstance(vec, DbDevExportInfos):
        raise TypeError("vec must be a tango.DbDevExportInfos")
    for e in seq:
        vec.append(e)
    return vec


def seq_2_DbData(seq, vec=None):
    """Converts a python sequence<DbDatum> object to a :class:`tango.DbData`

    :param seq: the sequence of DbDatum
    :type seq: sequence<DbDatum>
    :param vec: (optional, default is None) an :class:`tango.DbData`
                to be filled. If None is given, a new :class:`tango.DbData`
                is created
    :return: a :class:`tango.DbData` filled with the same contents as seq
    :rtype: :class:`tango.DbData`
    """
    if vec is None:
        if isinstance(seq, DbData):
            return seq
        vec = DbData()
    if not isinstance(vec, DbData):
        raise TypeError("vec must be a tango.DbData")
    else:
        for e in seq:
            if isinstance(e, DbDatum):
                vec.append(e)
            else:
                e = ensure_binary(e, "latin-1")
                vec.append(DbDatum(e))
    return vec


def DbData_2_dict(db_data, d=None):
    if d is None:
        d = {}
    if not isinstance(db_data, DbData):
        raise TypeError(
            f"db_data must be a tango.DbData. A {type(db_data)} found instead"
        )
    for db_datum in db_data:
        d[db_datum.name] = db_datum.value_string
    return d


def seqStr_2_obj(seq, tg_type, tg_format=None):
    """Translates a sequence<str> to a sequence of objects of give type and format

    :param seq: the sequence
    :type seq: sequence<str>
    :param tg_type: tango type
    :type tg_type: :class:`tango.CmdArgType`
    :param tg_format: (optional, default is None, meaning SCALAR) tango format
    :type tg_format: :class:`tango.AttrDataFormat`

    :return: a new sequence
    """
    if tg_format:
        return _seqStr_2_obj_from_type_format(seq, tg_type, tg_format)
    return _seqStr_2_obj_from_type(seq, tg_type)


def _seqStr_2_obj_from_type(seq, tg_type):
    if is_pure_str(seq):
        seq = (seq,)

    # Scalar cases
    global _scalar_int_types
    if tg_type in _scalar_int_types:
        return int(seq[0])

    global _scalar_float_types
    if tg_type in _scalar_float_types:
        return float(seq[0])

    global _scalar_str_types
    if tg_type in _scalar_str_types:
        return seq[0]

    if tg_type == CmdArgType.DevBoolean:
        return seq[0].lower() == "true"

    # sequence cases
    if tg_type in (CmdArgType.DevVarCharArray, CmdArgType.DevVarStringArray):
        return seq

    global _array_int_types
    if tg_type in _array_int_types:
        argout = []
        for x in seq:
            argout.append(int(x))
        return argout

    global _array_float_types
    if tg_type in _array_float_types:
        argout = []
        for x in seq:
            argout.append(float(x))
        return argout

    if tg_type == CmdArgType.DevVarBooleanArray:
        argout = []
        for x in seq:
            argout.append(x.lower() == "true")
        return argout

    return []


def _seqStr_2_obj_from_type_format(seq, tg_type, tg_format):
    if tg_format == AttrDataFormat.SCALAR:
        return _seqStr_2_obj_from_type(tg_type, seq)
    elif tg_format == AttrDataFormat.SPECTRUM:
        return _seqStr_2_obj_from_type(_scalar_to_array_type[tg_type], seq)
    elif tg_format == AttrDataFormat.IMAGE:
        if tg_type == CmdArgType.DevString:
            return seq

        global _scalar_int_types
        if tg_type in _scalar_int_types:
            argout = []
            for x in seq:
                tmp = []
                for y in x:
                    tmp.append(int(y))
                argout.append(tmp)
            return argout

        global _scalar_float_types
        if tg_type in _scalar_float_types:
            argout = []
            for x in seq:
                tmp = []
                for y in x:
                    tmp.append(float(y))
                argout.append(tmp)
            return argout

    # UNKNOWN_FORMAT
    return _seqStr_2_obj_from_type(tg_type, seq)


def scalar_to_array_type(tg_type):
    """
    Gives the array tango type corresponding to the given tango
    scalar type. Example: giving DevLong will return DevVarLongArray.

    :param tg_type: tango type
    :type tg_type: :class:`tango.CmdArgType`

    :return: the array tango type for the given scalar tango type
    :rtype: :class:`tango.CmdArgType`

    :raises ValueError: in case the given dtype is not a tango scalar type
    """
    try:
        return _scalar_to_array_type[tg_type]
    except KeyError:
        raise ValueError(f"Invalid tango scalar type: {tg_type}")


def str_2_obj(obj_str, tg_type=None):
    """Converts a string into an object according to the given tango type

    :param obj_str: the string to be converted
    :type obj_str: :py:obj:`str`
    :param tg_type: tango type
    :type tg_type: :class:`tango.CmdArgType`
    :return: an object calculated from the given string
    :rtype: :py:obj:`object`
    """
    if tg_type is None:
        return obj_str
    f = str
    if is_scalar_type(tg_type):
        if is_numerical_type(tg_type):
            if obj_str in __NO_STR_VALUE:
                return None
        if is_int_type(tg_type):
            f = int
        elif is_float_type(tg_type):
            f = float
        elif is_bool_type(tg_type):
            f = bool_
    return f(obj_str)


def obj_2_str(obj, tg_type=None):
    """Converts a python object into a string according to the given tango type

    :param obj: the object to be converted
    :type obj: :py:obj:`object`
    :param tg_type: tango type
    :type tg_type: :class:`tango.CmdArgType`
    :return: a string representation of the given object
    :rtype: :py:obj:`str`
    """
    if tg_type is None:
        return obj
    if tg_type in _scalar_types:
        # scalar cases
        if is_pure_str(obj):
            return obj
        elif is_non_str_seq(obj):
            if not len(obj):
                return ""
            obj = obj[0]
        return str(obj)
    # sequence cases
    if obj is None:
        return ""
    return "\n".join([str(i) for i in obj])


def _append_dict_to_db_data(db_data, value):
    for k, v in value.items():
        if isinstance(v, DbDatum):
            db_data.append(v)
            continue
        db_datum = DbDatum(k)
        if is_non_str_seq(v):
            seq_2_StdStringVector(v, db_datum.value_string)
        else:
            if not is_pure_str(v):
                v = str(v)
            v = ensure_binary(v, encoding="latin-1")
            db_datum.value_string.append(v)
        db_data.append(db_datum)


def parameter_2_dbdata(param, param_name):
    if isinstance(param, DbData):
        return param
    elif isinstance(param, DbDatum):
        new_param = DbData()
        new_param.append(param)
        return new_param
    elif is_pure_str(param):
        new_param = DbData()
        new_param.append(DbDatum(param))
        return new_param
    elif is_non_str_seq(param):
        return seq_2_DbData(param)
    elif isinstance(param, collections.abc.Mapping):
        new_param = DbData()
        if len(param) > 0:
            # for attributes we have nested dict
            first_dict_value = list(param.values())[0]
            if isinstance(first_dict_value, collections.abc.Mapping):
                for k1, v1 in param.items():
                    attr = DbDatum(k1)
                    attr.append(str(len(v1)))
                    new_param.append(attr)
                    _append_dict_to_db_data(new_param, v1)
            else:
                _append_dict_to_db_data(new_param, param)
        return new_param

    raise TypeError(
        f"{param_name} must be a str, tango.DbDatum, tango.DbData, "
        "a sequence<DbDatum>, a sequence<str> or a dictionary"
    )


def get_property_from_db(self, propname, user_dbdata):
    if user_dbdata is None:
        user_dbdata = DbData()
    if is_pure_str(propname):
        self._get_property(propname, user_dbdata)
        return DbData_2_dict(user_dbdata)
    elif isinstance(propname, (collections.abc.Sequence, StdStringVector)):
        if len(propname) == 0:
            return {}
        all_str = True
        for p in propname:
            if not is_pure_str(p):
                all_str = False
                break
        if all_str:
            self._get_property(propname, user_dbdata)
            return DbData_2_dict(user_dbdata)

    dbdata = parameter_2_dbdata(propname, "propname")
    if len(dbdata) == 0:
        return {}
    self._get_property(dbdata)
    return DbData_2_dict(dbdata)


def __get_meth_func(klass, method_name):
    meth = getattr(klass, method_name)
    func = meth
    if hasattr(meth, "__func__"):
        func = meth.__func__
    elif hasattr(meth, "im_func"):
        func = meth.im_func
    return meth, func


def copy_doc(klass, fnname):
    """Copies documentation string of a method from the super class into the
    rewritten method of the given class"""
    base_meth, base_func = __get_meth_func(klass.__base__, fnname)
    meth, func = __get_meth_func(klass, fnname)
    func.__doc__ = base_func.__doc__


def document_method(klass, method_name, d, add=True):
    meth, func = __get_meth_func(klass, method_name)
    if add:
        cpp_doc = meth.__doc__
        if cpp_doc:
            func.__doc__ = f"{d}\n{cpp_doc}"
            return
    func.__doc__ = d

    if func.__name__ != method_name:
        try:
            func.__name__ = method_name
        except AttributeError:
            pass


def document_static_method(klass, method_name, d, add=True):
    meth, func = __get_meth_func(klass, method_name)
    if add:
        cpp_doc = meth.__doc__
        if cpp_doc:
            meth.__doc__ = f"{d}\n{cpp_doc}"
            return
    meth.__doc__ = d


class CaselessList(list):
    """A case insensitive lists that has some caseless methods. Only allows
    strings as list members. Most methods that would normally return a list,
    return a CaselessList. (Except list() and lowercopy())
    Sequence Methods implemented are :
    __contains__, remove, count, index, append, extend, insert,
    __getitem__, __setitem__, __getslice__, __setslice__
    __add__, __radd__, __iadd__, __mul__, __rmul__
    Plus Extra methods:
    findentry, copy , lowercopy, list
    Inherited methods :
    __imul__, __len__, __iter__, pop, reverse, sort
    """

    def __init__(self, inlist=[]):
        list.__init__(self)
        for entry in inlist:
            if not isinstance(entry, str):
                raise TypeError(
                    f"Members of this object must be strings. "
                    f'You supplied "{entry}" which is "{type(entry)}"'
                )
            self.append(entry)

    def findentry(self, item):
        """A caseless way of checking if an item is in the list or not.
        It returns None or the entry."""
        if not isinstance(item, str):
            raise TypeError(
                f'Members of this object must be strings. You supplied "{type(item)}"'
            )
        for entry in self:
            if item.lower() == entry.lower():
                return entry
        return None

    def __contains__(self, item):
        """A caseless way of checking if a list has a member in it or not."""
        for entry in self:
            if item.lower() == entry.lower():
                return True
        return False

    def remove(self, item):
        """Remove the first occurence of an item, the caseless way."""
        for entry in self:
            if item.lower() == entry.lower():
                list.remove(self, entry)
                return
        raise ValueError(": list.remove(x): x not in list")

    def copy(self):
        """Return a CaselessList copy of self."""
        return CaselessList(self)

    def list(self):
        """Return a normal list version of self."""
        return list(self)

    def lowercopy(self):
        """Return a lowercase (list) copy of self."""
        return [entry.lower() for entry in self]

    def append(self, item):
        """Adds an item to the list and checks it's a string."""
        if not isinstance(item, str):
            raise TypeError(
                f'Members of this object must be strings. You supplied "{type(item)}"'
            )
        list.append(self, item)

    def extend(self, item):
        """Extend the list with another list. Each member of the list must be
        a string."""
        if not isinstance(item, list):
            raise TypeError(
                f'You can only extend lists with lists. You supplied "{type(item)}"'
            )
        for entry in item:
            if not isinstance(entry, str):
                raise TypeError(
                    f"Members of this object must be strings. "
                    f'You supplied "{type(entry)}"'
                )
            list.append(self, entry)

    def count(self, item):
        """Counts references to 'item' in a caseless manner.
        If item is not a string it will always return 0."""
        if not isinstance(item, str):
            return 0
        count = 0
        for entry in self:
            if item.lower() == entry.lower():
                count += 1
        return count

    def index(self, item, minindex=0, maxindex=None):
        """Provide an index of first occurence of item in the list. (or raise
        a ValueError if item not present)
        If item is not a string, will raise a TypeError.
        minindex and maxindex are also optional arguments
        s.index(x[, i[, j]]) return smallest k such that s[k] == x and i <= k < j
        """
        if maxindex is None:
            maxindex = len(self)
        minindex = max(0, minindex) - 1
        maxindex = min(len(self), maxindex)
        if not isinstance(item, str):
            raise TypeError(
                f'Members of this object must be strings. You supplied "{type(item)}"'
            )
        index = minindex
        while index < maxindex:
            index += 1
            if item.lower() == self[index].lower():
                return index
        raise ValueError(": list.index(x): x not in list")

    def insert(self, i, x):
        """s.insert(i, x) same as s[i:i] = [x]
        Raises TypeError if x isn't a string."""
        if not isinstance(x, str):
            raise TypeError(
                f'Members of this object must be strings. You supplied "{type(x)}"'
            )
        list.insert(self, i, x)

    def __setitem__(self, index, value):
        """For setting values in the list.
        index must be an integer or (extended) slice object. (__setslice__ used
        for simple slices)
        If index is an integer then value must be a string.
        If index is a slice object then value must be a list of strings - with
        the same length as the slice object requires.
        """
        if isinstance(index, int):
            if not isinstance(value, str):
                raise TypeError(
                    f"Members of this object must be strings. "
                    f'You supplied "{type(value)}"'
                )
            list.__setitem__(self, index, value)
        elif isinstance(index, slice):
            if not hasattr(value, "__len__"):
                raise TypeError("Value given to set slice is not a sequence object.")
            for entry in value:
                if not isinstance(entry, str):
                    raise TypeError(
                        f"Members of this object must be strings. "
                        f'You supplied "{type(entry)}"'
                    )
            list.__setitem__(self, index, value)
        else:
            raise TypeError("Indexes must be integers or slice objects.")

    def __setslice__(self, i, j, sequence):
        """Called to implement assignment to self[i:j]."""
        for entry in sequence:
            if not isinstance(entry, str):
                raise TypeError(
                    f"Members of this object must be strings. "
                    f'You supplied "{type(entry)}"'
                )
        list.__setslice__(self, i, j, sequence)

    def __getslice__(self, i, j):
        """Called to implement evaluation of self[i:j].
        Although the manual says this method is deprecated - if I don't define
        it the list one is called.
        (Which returns a list - this returns a CaselessList)"""
        return CaselessList(list.__getslice__(self, i, j))

    def __getitem__(self, index):
        """For fetching indexes.
        If a slice is fetched then the list returned is a CaselessList."""
        if not isinstance(index, slice):
            return list.__getitem__(self, index)
        else:
            return CaselessList(list.__getitem__(self, index))

    def __add__(self, item):
        """To add a list, and return a CaselessList.
        Every element of item must be a string."""
        return CaselessList(list.__add__(self, item))

    def __radd__(self, item):
        """To add a list, and return a CaselessList.
        Every element of item must be a string."""
        return CaselessList(list.__add__(self, item))

    def __iadd__(self, item):
        """To add a list in place."""
        for entry in item:
            self.append(entry)

    def __mul__(self, item):
        """To multiply itself, and return a CaselessList.
        Every element of item must be a string."""
        return CaselessList(list.__mul__(self, item))

    def __rmul__(self, item):
        """To multiply itself, and return a CaselessList.
        Every element of item must be a string."""
        return CaselessList(list.__rmul__(self, item))


class CaselessDict(dict):
    def __init__(self, other=None):
        if other:
            # Doesn't do keyword args
            if isinstance(other, dict):
                for k, v in other.items():
                    dict.__setitem__(self, k.lower(), v)
            else:
                for k, v in other:
                    dict.__setitem__(self, k.lower(), v)

    def __getitem__(self, key):
        return dict.__getitem__(self, key.lower())

    def __setitem__(self, key, value):
        dict.__setitem__(self, key.lower(), value)

    def __contains__(self, key):
        return dict.__contains__(self, key.lower())

    def __delitem__(self, k):
        dict.__delitem__(self, k.lower())

    def has_key(self, key):
        return key.lower() in self

    def get(self, key, def_val=None):
        return dict.get(self, key.lower(), def_val)

    def setdefault(self, key, def_val=None):
        return dict.setdefault(self, key.lower(), def_val)

    def update(self, other):
        for k, v in other.items():
            dict.__setitem__(self, k.lower(), v)

    def fromkeys(self, iterable, value=None):
        d = CaselessDict()
        for k in iterable:
            dict.__setitem__(d, k.lower(), value)
        return d

    def pop(self, key, def_val=None):
        return dict.pop(self, key.lower(), def_val)

    def keys(self):
        return CaselessList(dict.keys(self))


class EventCallback:
    """
    Useful event callback for test purposes

    Usage::

        >>> dev = tango.DeviceProxy(dev_name)
        >>> cb = tango.utils.EventCallback()
        >>> id = dev.subscribe_event("state", tango.EventType.CHANGE_EVENT, cb, [])
        2011-04-06 15:33:18.910474 sys/tg_test/1 STATE CHANGE [ATTR_VALID] ON

    Allowed format keys are:

        - date (event timestamp)
        - reception_date (event reception timestamp)
        - type (event type)
        - reason (event subscription reason)
        - dev_name (device name)
        - name (attribute name)
        - value (event value)

    New in PyTango 7.1.4
    """

    def __init__(
        self,
        format="{date} {dev_name} {name} {type} {reason} {value}",
        fd=sys.stdout,
        max_buf=100,
    ):
        self._msg = format
        self._fd = fd
        self._evts = []
        self._max_buf = max_buf

    def get_events(self):
        """Returns the list of events received by this callback

        :return: the list of events received by this callback
        :rtype: sequence<obj>
        """
        return self._evts

    def push_event(self, evt):
        """Internal usage only"""
        try:
            self._push_event(evt)
        except Exception as e:
            print(f"Unexpected error in callback for {evt}: {e}", file=self._fd)

    def _push_event(self, evt):
        """Internal usage only"""
        self._append(evt)
        import datetime

        now = datetime.datetime.now()

        try:
            date = self._get_date(evt)
        except Exception:
            date = now

        try:
            reception_date = evt.reception_date.todatetime()
        except Exception:
            reception_date = now

        try:
            evt_type = evt.event.upper()
        except Exception:
            evt_type = "<UNKNOWN>"

        try:
            evt_reason = str(evt.event_reason).upper()
        except Exception:
            evt_reason = "<UNKNOWN>"

        try:
            dev_name = evt.device.dev_name().upper()
        except Exception:
            dev_name = "<UNKNOWN>"

        try:
            if hasattr(evt, "attr_name"):
                attr_name = evt.attr_name.split("/")[-1].upper()
                attr_name = attr_name.removesuffix("#DBASE=NO")
            else:
                attr_name = "<N/A>"
        except Exception:
            attr_name = "<UNKNOWN>"

        try:
            value = self._get_value(evt)
        except Exception as e:
            value = f"Unexpected exception in getting event value: {e}"

        d = {
            "date": date,
            "reception_date": reception_date,
            "type": evt_type,
            "reason": evt_reason,
            "dev_name": dev_name,
            "name": attr_name,
            "value": value,
        }
        print(self._msg.format(**d), file=self._fd)

    def _append(self, evt):
        """Internal usage only"""
        evts = self._evts
        if len(evts) == self._max_buf:
            evts.pop(0)
        evts.append(evt)

    def _get_date(self, evt):
        if isinstance(evt, EventData):
            return evt.attr_value.time.todatetime()
        else:
            return evt.get_date().todatetime()

    def _get_value(self, evt):
        """Internal usage only"""
        if evt.err:
            e = evt.errors[0]
            return f"[{e.reason}] {e.desc}"

        if isinstance(evt, EventData):
            if evt.attr_value is None and evt.event_reason == EventReason.SubSuccess:
                return "<N/A>"  # assume EventSubMode.Async which provides no value
            else:
                return f"[{evt.attr_value.quality}] {evt.attr_value.value}"
        elif isinstance(evt, AttrConfEventData):
            cfg = evt.attr_conf
            return f"label='{cfg.label}'; unit='{cfg.unit}'"
        elif isinstance(evt, DataReadyEventData):
            return f"ctr={evt.ctr}"
        elif isinstance(evt, DevIntrChangeEventData):
            return (
                f"dev_started={evt.dev_started}; "
                f"{len(evt.cmd_list)} commands; {len(evt.att_list)} attrs"
            )


class AsyncEventCallback(EventCallback):
    async def push_event(self, evt):
        """Internal usage only"""
        try:
            self._push_event(evt)
        except Exception as e:
            print(f"Unexpected error in callback for {evt}: {e}", file=self._fd)


def get_home():
    """
    Find user's home directory if possible. Otherwise raise error.

    :return: user's home directory
    :rtype: :py:obj:`str`

    New in PyTango 7.1.4
    """
    path = ""
    try:
        path = os.path.expanduser("~")
    except Exception:
        pass
    if not os.path.isdir(path):
        for evar in ("HOME", "USERPROFILE", "TMP"):
            try:
                path = os.environ[evar]
                if os.path.isdir(path):
                    break
            except Exception:
                pass
    if path:
        return path
    else:
        raise RuntimeError("please define environment variable $HOME")


def _get_env_var(env_var_name):
    """
    Returns the value for the given environment name

    Search order:

        * a real environ var
        * HOME/.tangorc
        * /etc/tangorc

    :param env_var_name: the environment variable name
    :type env_var_name: str
    :return: the value for the given environment name
    :rtype: str

    New in PyTango 7.1.4
    """

    if env_var_name in os.environ:
        return os.environ[env_var_name]

    fname = os.path.join(get_home(), ".tangorc")
    if not os.path.exists(fname):
        if os.name == "posix":
            fname = "/etc/tangorc"
    if not os.path.exists(fname):
        return None

    for line in open(fname):
        strippedline = line.split("#", 1)[0].strip()

        if not strippedline:
            # empty line
            continue

        tup = strippedline.split("=", 1)
        if len(tup) != 2:
            # illegal line!
            continue

        key, val = map(str.strip, tup)
        if key == env_var_name:
            return val


def from_version_str_to_hex_str(version_str):
    v = map(int, version_str.split("."))
    return "0x%02d%02d%02d00" % (v[0], v[1], v[2])


def from_version_str_to_int(version_str):
    return int(from_version_str_to_hex_str(version_str), 16)


def info():
    # Compile and Runtime are set by `tango.pytango_init.init`
    from tango.constants import Compile, Runtime

    msg = f"""\
PyTango {Release.version_long} {Release.version_info}
PyTango compiled with:
    Python   : {Compile.PY_VERSION}
    Numpy    : {Compile.NUMPY_VERSION}
    Tango    : {Compile.TANGO_VERSION}
    pybind11 : {Compile.PYBIND11_VERSION}

PyTango runtime is:
    Python   : {Runtime.PY_VERSION}
    Numpy    : {Runtime.NUMPY_VERSION}
    Tango    : {Runtime.TANGO_VERSION}

PyTango running on:
{Runtime.UNAME}
"""
    return msg


def get_attrs(obj):
    """Helper for dir2 implementation."""
    if not hasattr(obj, "__dict__"):
        return []  # slots only
    proxy_type = types.MappingProxyType
    if not isinstance(obj.__dict__, (dict, proxy_type)):
        print(type(obj.__dict__), obj)
        raise TypeError(f"{obj.__name__}.__dict__ is not a dictionary")
    return obj.__dict__.keys()


def dir2(obj):
    """Default dir implementation.

    Inspired by gist: katyukha/dirmixin.py
    https://gist.github.com/katyukha/c6e5e2b829e247c9b009
    """
    attrs = set()

    if not hasattr(obj, "__bases__"):
        # obj is an instance
        if not hasattr(obj, "__class__"):
            # slots
            return sorted(get_attrs(obj))
        klass = obj.__class__
        attrs.update(get_attrs(klass))
    else:
        # obj is a class
        klass = obj

    for cls in klass.__bases__:
        attrs.update(get_attrs(cls))
        attrs.update(dir2(cls))
    attrs.update(get_attrs(obj))
    return list(attrs)


def ensure_binary(s, encoding="utf-8", errors="strict"):
    """Coerce **s** to the bytes type.
    For Python 3:
        - `str` -> encoded to `bytes`
        - `bytes` -> `bytes`

    Code taken from https://github.com/benjaminp/six/blob/1.12.0/six.py#L853
    """
    if isinstance(s, str):
        return s.encode(encoding, errors)
    elif isinstance(s, bytes):
        return s
    else:
        raise TypeError(f"not expecting type '{type(s)}'")


class PyTangoHelpFormatter(HelpFormatter):
    def _format_usage(self, usage, actions, groups, prefix):
        usage = super()._format_usage(usage, actions, groups, prefix)
        try:
            db = Database()
            servers_list = db.get_instance_name_list(self._prog)
            if servers_list.size():
                usage += (
                    f"Instance names defined in database for server {self._prog}:\n"
                )
                for server in servers_list:
                    usage += "\t" + str(server) + "\n"
            else:
                usage += f"Warning! No defined instance in database for server {self._prog} found!\n"
        except DevFailed:
            pass

        return usage


__TEST_CONTEXT_HOST_TRL = None


def _set_test_context_tango_host_fqtrl(host_trl):
    """For PyTango internal use only!"""
    # NOTE: we only keep one value so if multiple TestContexts are started in
    # the same process, only the latest one will be used
    if host_trl is not None:
        global __TEST_CONTEXT_HOST_TRL
        __TEST_CONTEXT_HOST_TRL = host_trl


def _clear_test_context_tango_host_fqtrl():
    """For PyTango internal use only!"""
    global __TEST_CONTEXT_HOST_TRL
    __TEST_CONTEXT_HOST_TRL = None


def _get_test_context_tango_host_fqtrl():
    """For PyTango internal use only!"""
    return __TEST_CONTEXT_HOST_TRL


def _get_device_fqtrl_if_necessary(device_trl):
    """For PyTango internal use only!"""
    if __TEST_CONTEXT_HOST_TRL:
        device_trl = _get_device_fqtrl(device_trl, __TEST_CONTEXT_HOST_TRL)
    return device_trl


def _get_device_fqtrl(device_trl, host_trl):
    parsed_device = urlparse(device_trl)
    if not _is_tango_uri_resolved(parsed_device):
        parsed_host = urlparse(host_trl)
        device_trl = _try_resolve_tango_trl(parsed_host, parsed_device)
    return device_trl


def _is_tango_uri_resolved(parsed_device):
    return parsed_device.scheme == "tango"


def _try_resolve_tango_trl(parsed_host, parsed_device):
    if not _is_valid_tango_trl(parsed_host):
        raise InvalidTangoHostTrlError(
            f"Invalid form for Tango host: {parsed_host!r}, device {parsed_device!r}. "
            f"(Override set to: {__TEST_CONTEXT_HOST_TRL})."
        )
    return _resolve_tango_trl(parsed_host, parsed_device)


def _is_valid_tango_trl(parsed_host):
    scheme_ok = parsed_host.scheme == "tango"
    hostname_ok = bool(parsed_host.hostname)
    port_ok = bool(parsed_host.port)
    path_ok = parsed_host.path == ""
    params_ok = parsed_host.params == ""
    query_ok = parsed_host.query == ""
    fragment_ok = parsed_host.fragment in ["", "dbase=no", "dbase=yes"]
    return (
        scheme_ok
        and hostname_ok
        and port_ok
        and path_ok
        and params_ok
        and query_ok
        and fragment_ok
    )


def _resolve_tango_trl(parsed_host, parsed_device):
    return urlunparse(
        [
            parsed_host.scheme,
            parsed_host.netloc,
            parsed_device.path,
            parsed_host.params,
            parsed_host.query,
            parsed_host.fragment,
        ]
    )


class InvalidTangoHostTrlError(ValueError):
    """Invalid Tango Resource Locator format for TANGO_HOST-like variable."""


class PyTangoUserWarning(UserWarning):
    # a custom category for all PyTango's warnings to give users the option of filtering PyTango's warnings
    pass


def _is_coroutine_function(obj):
    while isinstance(obj, functools.partial):
        obj = obj.func

    return inspect.iscoroutinefunction(obj) or (
        callable(obj) and inspect.iscoroutinefunction(obj.__call__)
    )


def _truthy_env_var(name) -> bool:
    value = ApiUtil.get_env_var(name)
    if value and value.lower() in {"on", "1", "true", "yes", "y"}:
        return True
    return False


_traced_coverage_run_active = False

try:
    import coverage

    _coverage = coverage.Coverage.current()
    if _coverage:
        _coverage_core = dict(_coverage.sys_info()).get("core", "").lower()
        if _coverage_core in {"pytracer", "ctracer"}:
            if _truthy_env_var("PYTANGO_DISABLE_COVERAGE_TRACE_PATCHING"):
                warnings.warn(
                    "Coverage run detected, but PYTANGO_DISABLE_COVERAGE_TRACE_PATCHING "
                    "environment variable is set. Reported coverage may be inaccurate.",
                    category=PyTangoUserWarning,
                )
            else:
                if getattr(threading, "_trace_hook", None):
                    _traced_coverage_run_active = True
                    warnings.warn(
                        "Coverage run detected: tango.server.Device methods "
                        "will be patched for tracing.",
                        category=PyTangoUserWarning,
                    )
                else:
                    warnings.warn(
                        "Coverage run detected, but unable to get threading._trace_hook. "
                        "Reported coverage may be inaccurate.",
                        category=PyTangoUserWarning,
                    )
        # (else, using sys.monitoring hooks or not tracing, so patching not required)

except Exception:
    pass

_traced_debug_run_active = False
pydevd = None

try:
    _disabled_via_env_var = _truthy_env_var("PYTANGO_DISABLE_DEBUG_TRACE_PATCHING")
    if not _disabled_via_env_var:
        _forced_via_env_var = _truthy_env_var("PYTANGO_FORCE_DEBUG_TRACE_PATCHING")
        if sys.version_info < (3, 12) or _forced_via_env_var:
            if "PYDEVD_DISABLE_FILE_VALIDATION" not in os.environ:
                os.environ["PYDEVD_DISABLE_FILE_VALIDATION"] = "1"
            import pydevd

            _debugger = pydevd.get_global_debugger()
        else:
            # we assume debugger using sys.monitoring hooks so it doesn't need patching
            _debugger = None

        if _debugger is not None:
            if _traced_coverage_run_active:
                warnings.warn(
                    "Debugger detected, but coverage run also detected. "
                    "Patching only for coverage, not for debugger.",
                    category=PyTangoUserWarning,
                )
            else:
                _traced_debug_run_active = True
                warnings.warn(
                    "Debugger detected: tango.server.Device methods "
                    "will be patched for tracing.",
                    category=PyTangoUserWarning,
                )
except Exception:
    pass


_telemetry_active = False
try:
    _globally_enabled_via_env_var = _truthy_env_var("TANGO_TELEMETRY_ENABLE")
    _locally_disabled_via_env_var = _truthy_env_var(
        "PYTANGO_DISABLE_TELEMETRY_PATCHING"
    )
    if (
        _globally_enabled_via_env_var
        and not _locally_disabled_via_env_var
        and TELEMETRY_SUPPORTED
    ):
        from opentelemetry import trace as trace_api
        from opentelemetry import context as context_api
        from opentelemetry.trace.propagation.tracecontext import (
            TraceContextTextMapPropagator,
        )

        _traces_exporter_type = ApiUtil.get_env_var("TANGO_TELEMETRY_TRACES_EXPORTER")
        if not _traces_exporter_type:
            _traces_exporter_type = "console"
        _traces_exporter_endpoint = ApiUtil.get_env_var(
            "TANGO_TELEMETRY_TRACES_ENDPOINT"
        )

        _telemetry_sdk_available = False
        try:
            _traces_exporter_kwargs = {}
            if _traces_exporter_type.lower() == "grpc":
                from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import (
                    OTLPSpanExporter as SpanExporter,
                )

                if _traces_exporter_endpoint:
                    _traces_exporter_endpoint = _traces_exporter_endpoint.lower()
                    if _traces_exporter_endpoint.startswith("grpc://"):
                        # cppTango requires endpoints starting "grpc://" for gRPC, even
                        # though this is a non-standard scheme.  We convert to
                        # the more standard http:// for the OTel exporter library.
                        _traces_exporter_endpoint = _traces_exporter_endpoint.replace(
                            "grpc://", "http://"
                        )
                    _traces_exporter_kwargs = {"endpoint": _traces_exporter_endpoint}
            elif _traces_exporter_type.lower() == "http":
                from opentelemetry.exporter.otlp.proto.http.trace_exporter import (
                    OTLPSpanExporter as SpanExporter,
                )

                if _traces_exporter_endpoint:
                    _traces_exporter_kwargs = {"endpoint": _traces_exporter_endpoint}
            elif _traces_exporter_type.lower() == "none":
                SpanExporter = None
            else:
                from opentelemetry.sdk.trace.export import (
                    ConsoleSpanExporter as SpanExporter,
                )

                if _traces_exporter_endpoint:
                    if _traces_exporter_endpoint.lower() == "cerr":
                        _traces_exporter_kwargs = {"out": sys.stderr}
                    else:
                        _traces_exporter_kwargs = {"out": sys.stdout}

                if _traces_exporter_type.lower() != "console":
                    warnings.warn(
                        f"Unknown value '{_traces_exporter_type}' for "
                        f"TANGO_TELEMETRY_TRACES_EXPORTER. Options are: "
                        f"'console', 'grpc', 'http' and 'none'. Defaulting to 'console'."
                    )
                    _traces_exporter_type = "console"

            from opentelemetry.sdk.resources import (
                HOST_NAME,
                SERVICE_INSTANCE_ID,
                SERVICE_NAME,
                SERVICE_NAMESPACE,
                Resource,
            )
            from opentelemetry.sdk.trace import TracerProvider
            from opentelemetry.sdk.trace.export import (
                BatchSpanProcessor,
                SimpleSpanProcessor,
            )

            _telemetry_sdk_available = True
        except ImportError:
            warnings.warn(
                "OpenTelemetry SDK packages not available: "
                "no telemetry traces will be emitted from this device/client.\n"
                "To emit telemetry, install the SDK packages: "
                "opentelemetry-sdk and either:\n"
                "  opentelemetry-exporter-otlp-proto-grpc (for traces via gRPC), or\n"
                "  opentelemetry-exporter-otlp-proto-http (for traces via HTTP).\n"
                "This warning can be disabled.\n"
                "  Either set environment variable PYTANGO_DISABLE_TELEMETRY_PATCHING=on"
                " to disable PyTango's usage of telemetry.\n"
                "  Or, set the environment variable TANGO_TELEMETRY_ENABLE=off to "
                "disable all telemetry in cppTango and PyTango.",
                category=PyTangoUserWarning,
            )

        def _get_current_otel_context() -> context_api.Context:
            return context_api.get_current()

        @contextlib.contextmanager
        def _span_from_cpptango(
            device: "tango.server.Device", fn: callable  # noqa: F821
        ) -> typing.Iterator[trace_api.Span]:
            fn = inspect.unwrap(fn)
            if not hasattr(fn, "__code__") and hasattr(fn, "__call__"):
                fn = fn.__call__
            name = getattr(fn, "__qualname__", getattr(fn, "__name__", "unknown"))
            code = getattr(fn, "__code__", None)
            filepath = getattr(code, "co_filename", "unknown")
            lineno = getattr(code, "co_firstlineno", 0)

            carrier = _telemetry.get_trace_context()
            ctx = TraceContextTextMapPropagator().extract(carrier=carrier)
            device_tracer = device.get_telemetry_tracer()
            token = _current_telemetry_tracer.set(device_tracer)
            try:
                with device_tracer.start_as_current_span(
                    name, context=ctx, kind=trace_api.SpanKind.SERVER
                ) as span:
                    span.set_attribute("code.filepath", filepath)
                    span.set_attribute("code.lineno", lineno)
                    current_thread = threading.current_thread()
                    span.set_attribute("thread.id", hex(current_thread.ident))
                    span.set_attribute("thread.name", current_thread.name)
                    _add_client_ident_info(device, span)
                    yield span
            finally:
                _current_telemetry_tracer.reset(token)

        # get process ID from a string like "jive3.MainPanel - PID=43281"
        _PID_PATTERN = re.compile(r"\bPID=(\d+)\b", re.IGNORECASE)

        def _add_client_ident_info(device, span):
            if not is_omni_thread():  # get_client_ident may crash
                return
            ident = device.get_client_ident()
            if not ident:  # not from external or collocated client (could be polling)
                return
            if ident.client_lang in {LockerLanguage.JAVA, LockerLanguage.JAVA_6}:
                span.set_attribute(
                    "tango.client_ident.java_ident",
                    f"{ident.java_ident[0]:x}{ident.java_ident[1]:x}",
                )
                span.set_attribute(
                    "tango.client_ident.java_main_class", ident.java_main_class
                )
                match = _PID_PATTERN.search(ident.java_main_class)
                client_pid = int(match.group(1)) if match else 0
            else:
                client_pid = ident.client_pid
            span.set_attribute("tango.client_ident.location", ident.client_ip)
            span.set_attribute("tango.client_ident.pid", client_pid)
            span.set_attribute("tango.client_ident.lang", str(ident.client_lang))

        @contextlib.contextmanager
        def _span_to_cpptango(name: str):
            # use propagator to get W3C strings from active Python OpenTelemetry context
            carrier = {}
            TraceContextTextMapPropagator().inject(carrier)
            traceparent = carrier.get("traceparent", "")
            tracestate = carrier.get("tracestate", "")
            # create C++ TraceContextScope and set context from Python context
            with _telemetry.TraceContextScope(name, traceparent, tracestate):
                yield

        def _default_telemetry_tracer_provider_factory(
            service_name: str,
            service_instance_id: typing.Union[None, str] = None,
            extra_resource_attributes: typing.Union[None, dict[str, str]] = None,
        ) -> trace_api.TracerProvider:
            """Create default telemetry TracerProvider for a device.

            A TraceProvider is not used directly, but rather used to create a Tracer.

            See also OpenTelemetry's OTEL_EXPERIMENTAL_RESOURCE_DETECTORS environment
            variable, and other resource detectors. It may be possible to add additional
            information just using this environment variable.
            """
            if _telemetry_sdk_available and SpanExporter is not None:
                resource_attributes = {
                    HOST_NAME: socket.getfqdn(),
                    SERVICE_NAMESPACE: "tango",
                    SERVICE_NAME: service_name,
                }
                if service_instance_id:
                    resource_attributes[SERVICE_INSTANCE_ID] = service_instance_id
                if extra_resource_attributes:
                    resource_attributes.update(extra_resource_attributes)
                tracer_provider = TracerProvider(
                    resource=Resource.create(resource_attributes)
                )
                exporter = SpanExporter(**_traces_exporter_kwargs)
                processor = _create_span_processor(exporter)
                tracer_provider.add_span_processor(processor)
            else:
                tracer_provider = trace_api.NoOpTracerProvider()
            return tracer_provider

        def _create_span_processor(exporter):
            processor_type = ApiUtil.get_env_var(
                "PYTANGO_TELEMETRY_SPAN_PROCESSOR_TYPE"
            )
            if processor_type and processor_type.lower() == "simple":
                processor_class = SimpleSpanProcessor
            elif processor_type and processor_type.lower() == "batch":
                processor_class = BatchSpanProcessor
            else:
                if _traces_exporter_type.lower() == "console":
                    processor_class = SimpleSpanProcessor
                else:
                    processor_class = BatchSpanProcessor
            processor = processor_class(exporter)
            return processor

        def _create_device_telemetry_tracer(
            tracer_provider: trace_api.TracerProvider,
        ) -> trace_api.Tracer:
            """Create a standard telemetry Tracer for a device."""
            # See the following link for details on tracer naming:
            #   https://github.com/open-telemetry/opentelemetry-specification/blob/main/specification/trace/api.md#get-a-tracer
            # cppTango uses "tango.cpp".  We use "tango.python.server"
            device_tracer = trace_api.get_tracer(
                instrumenting_module_name="tango.python.server",
                instrumenting_library_version=Release.version,
                tracer_provider=tracer_provider,
            )
            return device_tracer

        def _create_client_telemetry_tracer() -> trace_api.Tracer:
            service_name = ApiUtil.get_env_var("PYTANGO_TELEMETRY_CLIENT_SERVICE_NAME")
            if not service_name:
                service_name = "pytango.client"
            tracer_provider_factory = get_telemetry_tracer_provider_factory()
            tracer_provider = tracer_provider_factory(service_name)
            # See comment in _create_device_telemetry_tracer about naming.
            # We differentiate between client and server tracers.
            tracer = trace_api.get_tracer(
                instrumenting_module_name="tango.python.client",
                instrumenting_library_version=Release.version,
                tracer_provider=tracer_provider,
            )
            return tracer

        @atexit.register
        def _shutdown_telemetry():
            _telemetry.cleanup_default_telemetry_interface()

        def _set_telemetry_sdk_log_level():
            level_str = ApiUtil.get_env_var("PYTANGO_TELEMETRY_SDK_LOG_LEVEL")
            if level_str:
                level_int = logging.getLevelNamesMapping()[level_str.upper()]
                names_csv = ApiUtil.get_env_var("PYTANGO_TELEMETRY_SDK_LOGGER_NAMES")
                if names_csv:
                    names = names_csv.split(",")
                else:
                    names = [
                        "opentelemetry.sdk.trace.export",
                        "opentelemetry.sdk._shared_internal",
                        "opentelemetry.exporter.otlp.proto.grpc.exporter",
                        "opentelemetry.exporter.otlp.proto.http._log_exporter",
                        "opentelemetry.exporter.otlp.proto.http.log_exporter",
                        "opentelemetry.exporter.otlp.proto.http.metric_exporter",
                        "opentelemetry.exporter.otlp.proto.http.trace_exporter",
                    ]
                for name in names:
                    logging.getLogger(name).setLevel(level_int)
                _telemetry.set_log_level(level_str)

        _set_telemetry_sdk_log_level()

        _telemetry_client_tracer: typing.Union[None, trace_api.Tracer] = None
        _telemetry_active = True
except ImportError:
    _skip_warning_during_own_tests = os.environ.get("PYTANGO_TESTS_RUNNING") == "True"
    if not _skip_warning_during_own_tests:
        warnings.warn(
            "\nOpenTelemetry packages not available: \n"
            "telemetry context will not be passed to other Tango devices, "
            "and no telemetry will be emitted from this device/client.\n"
            "To pass through telemetry context, install the API packages: "
            "opentelemetry-api\n"
            "To emit telemetry, install the SDK packages: "
            "opentelemetry-sdk and either:\n"
            "  opentelemetry-exporter-otlp-proto-grpc (for traces via gRPC), or\n"
            "  opentelemetry-exporter-otlp-proto-http (for traces via HTTP).\n"
            "This warning can be disabled:\n"
            "  Either set environment variable PYTANGO_DISABLE_TELEMETRY_PATCHING=on "
            "to disable PyTango's usage of telemetry.\n"
            "  Or, set the environment variable TANGO_TELEMETRY_ENABLE=off to disable "
            "all telemetry in cppTango and PyTango.",
            category=PyTangoUserWarning,
        )
except Exception as exc:
    warnings.warn(
        f"Error setting up telemetry. Telemetry context may not be passed on "
        f"and traces may not be emitted. Possibly a PyTango bug.\n"
        f"Error: {exc!r}"
    )


_DummySpanContext = namedtuple(
    "_DummySpanContext",
    ["trace_id", "span_id", "is_remote", "trace_flags", "trace_state", "is_valid"],
)


class _DummySpan:
    def set_attributes(self, *args, **kwargs):
        pass

    def set_attribute(self, *args, **kwargs):
        pass

    def add_event(self, *args, **kwargs):
        pass

    def add_link(self, *args, **kwargs):
        pass

    def update_name(self, *args, **kwargs):
        pass

    def is_recording(self):
        return False

    def set_status(self, *args, **kwargs):
        pass

    def record_exception(self, *args, **kwargs):
        pass

    def end(self, *args, **kwargs):
        pass

    def get_span_context(self):
        return _DummySpanContext(0, 0, False, 0, {}, False)

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.end()


class _DummyTracer:
    def start_span(self, *args, **kwargs):
        pass

    @contextlib.contextmanager
    def start_as_current_span(self, *args, **kwargs):
        yield _DummySpan()


class _DummyTracerProvider:
    def get_tracer(self, *args, **kwargs):
        return _DummyTracer()


if not _telemetry_active:
    # define dummy handlers

    def _get_current_otel_context():
        return {}

    @contextlib.contextmanager
    def _span_to_cpptango(name: str):
        yield

    def _default_telemetry_tracer_provider_factory(
        service_name, service_instance_id=None, extra_resource_attributes=None
    ):
        return _DummyTracerProvider()

    def _create_device_telemetry_tracer(tracer_provider) -> _DummyTracer:
        return _DummyTracer()

    _telemetry_client_tracer = None

_current_telemetry_tracer = ContextVar("current_telemetry_tracer")

_current_telemetry_tracer_provider_factory = _default_telemetry_tracer_provider_factory


class _TracerProviderFactory(typing.Protocol):
    def __call__(
        self,
        service_name: str,
        service_instance_id: typing.Union[None, str] = None,
        extra_resource_attributes: typing.Union[None, dict[str, str]] = None,
    ) -> "opentelemetry.trace.TracerProvider":  # noqa: F821
        ...


def set_telemetry_tracer_provider_factory(provider_factory: _TracerProviderFactory):
    """Change the factory that will be used to create tracer providers.

    The factory is called when a tracer provider needs to be created.
    I.e., once for client access, and once for each device.
    """
    global _current_telemetry_tracer_provider_factory
    _current_telemetry_tracer_provider_factory = provider_factory


def get_telemetry_tracer_provider_factory() -> _TracerProviderFactory:
    """Get the factory that will be used to create tracer providers."""
    return _current_telemetry_tracer_provider_factory


_force_tracing = (
    _traced_debug_run_active or _traced_coverage_run_active or _telemetry_active
)


def _forcefully_traced_method(fn, is_kernel_method=False):
    # late import to avoid circular reference
    from tango.server import BaseDevice

    unwrapped_fn = inspect.unwrap(fn)

    def _get_device_telemetry_required(*args):
        device = None
        telemetry_required = False
        if _telemetry_active and args:
            first = args[0]
            if isinstance(first, BaseDevice):
                device = first
            else:
                fn_self = getattr(unwrapped_fn, "__self__", None)
                if isinstance(fn_self, BaseDevice):
                    device = fn_self
            if device is not None:
                telemetry_required = device.is_telemetry_enabled() and (
                    not is_kernel_method or device.is_kernel_tracing_enabled()
                )
        return device, telemetry_required

    def _set_sys_tracer_and_get_original():
        original_sys_tracer = "EMPTY"

        if _traced_coverage_run_active:
            original_sys_tracer = sys.gettrace()
            threading_trace_hook = getattr(threading, "_trace_hook", None)
            if threading_trace_hook:
                sys.settrace(threading_trace_hook)
        elif _traced_debug_run_active and pydevd is not None:
            pydevd.settrace(suspend=False, trace_only_current_thread=True)

        return original_sys_tracer

    @functools.wraps(fn)
    def trace_wrapper(*args, **kwargs):
        device, telemetry_required = _get_device_telemetry_required(*args)
        original_sys_tracer = _set_sys_tracer_and_get_original()
        try:
            if telemetry_required and device is not None:
                with _span_from_cpptango(device, fn):
                    ret = fn(*args, **kwargs)
            else:
                ret = fn(*args, **kwargs)
        finally:
            if original_sys_tracer != "EMPTY":
                sys.settrace(original_sys_tracer)
        return ret

    @functools.wraps(fn)
    async def async_trace_wrapper(*args, **kwargs):
        device, telemetry_required = _get_device_telemetry_required(*args)
        original_sys_tracer = _set_sys_tracer_and_get_original()
        try:
            if telemetry_required and device is not None:
                with _span_from_cpptango(device, fn):
                    ret = await fn(*args, **kwargs)
            else:
                ret = await fn(*args, **kwargs)
        finally:
            if original_sys_tracer != "EMPTY":
                sys.settrace(original_sys_tracer)
        return ret

    if _is_coroutine_function(fn):
        return async_trace_wrapper
    else:
        return trace_wrapper


def _trace_client(fn):
    """Wrapper/decorator to trace a client function for telemetry."""

    if _telemetry_active:
        # Change function names like "__DeviceProxy__subscribe_event" to
        # "DeviceProxy.subscribe_event" for better readability
        fn_name = getattr(fn, "__qualname__", getattr(fn, "__name__", "unknown"))
        match = re.match(r"__(?P<prefix>\w+?)__(?P<suffix>.*)", fn_name)
        if match:
            fn_name = f"{match.group('prefix')}.{match.group('suffix')}"

        @functools.wraps(fn)
        def client_trace_wrapper(*args, **kwargs):
            global _telemetry_client_tracer
            if _telemetry_client_tracer is None:
                _telemetry_client_tracer = _create_client_telemetry_tracer()
            tracer = _current_telemetry_tracer.get(_telemetry_client_tracer)
            location = kwargs.pop("trace_location", None)
            context = kwargs.pop("trace_context", None)
            if location is None:
                filename, lineno, qualname = _get_non_tango_source_location()
            else:
                filename, lineno, qualname = location
            with tracer.start_as_current_span(
                qualname, kind=trace_api.SpanKind.CLIENT, context=context
            ) as span:
                span.set_attribute("code.filepath", filename)
                span.set_attribute("code.lineno", lineno)
                current_thread = threading.current_thread()
                span.set_attribute("thread.id", hex(current_thread.ident))
                span.set_attribute("thread.name", current_thread.name)
                with _span_to_cpptango(fn_name):
                    return fn(*args, **kwargs)

        client_trace_wrapper.__signature__ = inspect.signature(fn)
        client_trace_wrapper.__trace_kwargs__ = True

    else:
        client_trace_wrapper = fn

    return client_trace_wrapper


_SourceLocation = namedtuple("_SourceLocation", ("filepath", "lineno", "qualname"))


def _get_non_tango_source_location(
    source: typing.Union[callable, None] = None
) -> _SourceLocation:
    """Provides non-PyTango source caller for logging and tracing functions.

    :param source:
        (optional) Method or function, which will be unwrapped of decorated wrappers
        and inspected for location. If not provided - current stack will be used to
        deduce the location.
    :type source: Callable

    :return:
        Named tuple (filepath, lineno, qualname)
    :rtype :_SourceLocation:
    """
    try:
        if source:
            # If source is a wrapped function - unwrap it to inner function
            source = inspect.unwrap(source)
            # Find callable code location
            code = getattr(source, "__code__", None)
            if code:
                filepath = code.co_filename
                lineno = code.co_firstlineno
                qualname = getattr(
                    code, "co_qualname", getattr(code, "co_name", str(source))
                )
                return _SourceLocation(filepath, lineno, qualname)
        else:
            caller, module = _get_first_non_tango_caller_and_module()
            if caller:
                code = caller.f_code
                filepath = code.co_filename
                lineno = caller.f_lineno
                qualname = getattr(
                    code,
                    "co_qualname",
                    getattr(code, "co_name", "unknown"),
                )
                if qualname == "<module>" and module in ("__main__", "__mp_main__"):
                    qualname = module
                return _SourceLocation(filepath, lineno, qualname)
        return _SourceLocation("(unknown)", 0, str(source))
    except Exception:
        return _SourceLocation("(unknown)", 0, str(source))


# Search the call stack until we are out of the 'tango' module. We cannot
# have a fixed number here because loggers/streams/tracing is used in many places
# inside pytango with varying call stack depth.
#
# There are different option below which trade compatibility for speed (fastest first).
if hasattr(sys, "_getframemodulename") and hasattr(sys, "_getframe"):

    def _get_first_non_tango_caller_and_module():
        depth = 2
        caller = None
        while True:
            module = sys._getframemodulename(depth)  # added in Python 3.12
            if module != "tango" and not module.startswith("tango."):
                caller = sys._getframe(depth)
                break
            elif module is None:
                break  # stack exhausted
            depth += 1
        if caller:
            return caller, module
        else:
            return None, ""

elif hasattr(sys, "_getframe"):

    def _get_first_non_tango_caller_and_module():
        depth = 2
        caller = None
        module = ""
        try:
            while True:
                caller = sys._getframe(depth)
                module = caller.f_globals["__name__"]
                if module != "tango" and not module.startswith("tango."):
                    break
                depth += 1
        except ValueError:
            pass  # stack exhausted
        if caller:
            return caller, module
        else:
            return None, ""

else:
    # use inspect (the slowest approach, most portable)
    def _get_first_non_tango_caller_and_module():
        for caller, _, _, _, _, _ in inspect.stack(0):
            module = caller.f_globals["__name__"]
            if module != "tango" and not module.startswith("tango."):
                return caller, module
        return None, ""


# Since cppTango based on omniThreads, we have to slightly modify ThreadPoolExecutor,
# to use threads with ensured dummy omniORB ID
#
# The `_adjust_thread_count` method below is a slightly modified version of the code from
# https://github.com/python/cpython/blob/3.12/Lib/concurrent/futures/thread.py
# and
# https://github.com/python/cpython/blob/3.14/Lib/concurrent/futures/thread.py

# Copyright (c) Python Software Foundation; All Rights Reserved

from concurrent.futures import ThreadPoolExecutor
from concurrent.futures.thread import _worker, _threads_queues


class PyTangoThreadPoolExecutor(ThreadPoolExecutor):
    """
    Based on concurrent.futures.ThreadPoolExecutor, but additionally ensures dummy omniORB runs for every running thread.
    """

    def _adjust_thread_count(self):
        # if idle threads are available, don't spin new threads
        if self._idle_semaphore.acquire(timeout=0):
            return

        # When the executor gets lost, the weakref callback will wake up
        # the worker threads.
        def weakref_cb(_, q=self._work_queue):
            q.put(None)

        num_threads = len(self._threads)
        if num_threads < self._max_workers:
            thread_name = "%s_%d" % (self._thread_name_prefix or self, num_threads)
            if hasattr(self, "_initializer"):
                # Python < 3.14
                t = threading.Thread(
                    name=thread_name,
                    target=_thread_pool_executor_worker,
                    args=(
                        weakref.ref(self, weakref_cb),
                        self._work_queue,
                        self._initializer,
                        self._initargs,
                    ),
                )
            else:
                # Python >= 3.14
                t = threading.Thread(
                    name=thread_name,
                    target=_thread_pool_executor_ctx_worker,
                    args=(
                        weakref.ref(self, weakref_cb),
                        self._create_worker_context(),
                        self._work_queue,
                    ),
                )
            t.start()
            self._threads.add(t)
            _threads_queues[t] = self._work_queue


# Override _worker for Python < 3.14
def _thread_pool_executor_worker(executor_reference, work_queue, initializer, initargs):
    with EnsureOmniThread():
        _worker(executor_reference, work_queue, initializer, initargs)


# Override _worker for Python >= 3.14
def _thread_pool_executor_ctx_worker(executor_reference, ctx, work_queue):
    with EnsureOmniThread():
        _worker(executor_reference, ctx, work_queue)


def _get_new_CallbackAutoDie():
    id = time.time_ns()
    cb = __CallBackAutoDie(id)
    __auto_die_callbacks_holder[id] = cb
    return cb


def _release_CallbackAutoDie(id):
    if id in __auto_die_callbacks_holder:
        del __auto_die_callbacks_holder[id]


def _get_command_inout_param(self, cmd_name, cmd_param=None):
    if cmd_param is None:
        return DeviceData()

    if isinstance(cmd_param, DeviceData):
        return cmd_param

    if isinstance(self, DeviceProxy):
        # This is not part of 'Connection' interface, but
        # DeviceProxy only.
        info = self.command_query(cmd_name)
        param = DeviceData()
        try:
            param.insert(info.in_type, cmd_param)
        except TypeError as err:
            raise TypeError(
                f"Invalid input argument for command {cmd_name}: "
                f"{cmd_param!r} cannot be converted to type {info.in_type}"
            ) from err
        return param
    elif isinstance(self, __Group):

        if self.get_size() == 0:
            return DeviceData()

        if isinstance(cmd_param, DeviceDataList):
            return cmd_param

        last_cause = None
        try:
            types = set()
            typ = None
            for idx in range(1, self.get_size() + 1):
                dev = self.get_device(idx)
                try:
                    typ = dev.command_query(cmd_name).in_type
                    types.add(typ)
                except DevFailed as df:
                    last_cause = df
            if not types:
                if last_cause:
                    Except.re_throw_exception(
                        last_cause,
                        "PyAPI_GroupCommandArgInTypeUnknown",
                        "Cannot fetch at least one command type in group.",
                        "Group.command_inout_asynch",
                    )
                else:
                    Except.throw_exception(
                        "PyAPI_GroupCommandArgInTypeUnknown",
                        "Cannot fetch at least one command type in group. Unknown cause.",
                        "Group.command_inout_asynch",
                    )
            elif len(types) > 1:
                raise TypeError(
                    "Cannot execute command with more than one type in group, types are:\n"
                    f"{types}"
                )
        finally:
            del last_cause

        param = DeviceData()
        try:
            param.insert(typ, cmd_param)
        except TypeError as err:
            raise TypeError(
                f"Invalid input argument for command {cmd_name}: "
                f"{cmd_param!r} cannot be converted to type {typ}"
            ) from err
        return param

    elif isinstance(self, Database):
        # I just try to guess types DevString and DevVarStringArray
        # as they are used for Database
        param = DeviceData()
        if isinstance(cmd_param, str):
            param.insert(CmdArgType.DevString, cmd_param)
            return param
        elif isinstance(cmd_param, collections.abc.Sequence) and all(
            [isinstance(x, str) for x in cmd_param]
        ):
            param.insert(CmdArgType.DevVarStringArray, cmd_param)
            return param
        else:
            raise TypeError(
                "command_inout() parameter must be a DeviceData object or a string or a sequence of strings"
            )
    else:
        raise TypeError("command_inout() parameter must be a DeviceData object.")


def _exception_converter(exception):

    # if user managed to create DevFailed, we do not need to convert it
    if isinstance(exception, DevFailed):
        return exception
    else:
        if exception.__traceback__ is None:
            # to generate DevFailed we need traceback
            # if user does not provide one (Exception.with_traceback), we generate our
            try:
                raise Exception()
            except Exception:
                # to get to the frame, where user called push_event
                traceback = sys.exc_info()[2]
                try:
                    user_frame = traceback.tb_frame.f_back.f_back
                    exception.__traceback__ = types.TracebackType(
                        None, user_frame, user_frame.f_lasti, user_frame.f_lineno
                    )
                except Exception:
                    # if fails, use what we have
                    try:
                        Except.throw_exception(
                            reason="PyDs_PythonError",
                            description=repr(exception),
                            origin="UNKNOWN: cannot get Python's traceback",
                            severity=ErrSeverity.ERR,
                        )
                    except DevFailed as err:
                        return err
        return Except.to_dev_failed(
            exception.__class__, exception, exception.__traceback__
        )


class _InterfaceDefinedByIDL:
    _initialized = False

    def __setattr__(self, name, value):
        if name not in self.__dict__ and self._initialized:
            raise AttributeError(
                f"tango._tango.{self.__class__.__name__} "
                f"object has no attribute '{name}'"
            )

        return super().__setattr__(name, value)


def _check_only_allowed_kwargs(call_kwargs, allowed_keys):
    unexpected_keys = call_kwargs.keys() - allowed_keys
    if unexpected_keys:
        raise TypeError(
            f"Got unexpected keyword argument(s): {', '.join(unexpected_keys)}.\n"
            f"Allowed keys are: {', '.join(sorted(allowed_keys))}"
        )


# Helper class to replace deprecated object
class _RemovedClass:
    def __init__(self, message):
        self._message = message

    def _trigger_error(self, *args, **kwargs):
        raise AttributeError(self._message)

    __getattr__ = _trigger_error
    __call__ = _trigger_error
    __getitem__ = _trigger_error



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
from tango import DevState
from tango.server import Device, device_property


class MandatoryPropertyServer(Device):
    Hostname = device_property(
        dtype="str", mandatory=True, doc="The controller host address"
    )

    Port = device_property(
        dtype="int", default_value=3456, doc="The controller port number"
    )

    def init_device(self):
        Device.init_device(self)
        print("Port: ", self.Port)
        print("Host: ", self.Hostname)
        self.set_state(DevState.ON)


if __name__ == "__main__":
    MandatoryPropertyServer.run_server()



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
"""Load tango-specific pytest fixtures."""

import multiprocessing
import sys
import os
import json
import shutil
from functools import partial
from subprocess import Popen

import pytest

from tango import DeviceProxy, GreenMode, Util
from tango.asyncio import DeviceProxy as asyncio_DeviceProxy
from tango.gevent import DeviceProxy as gevent_DeviceProxy
from tango.futures import DeviceProxy as futures_DeviceProxy
from tango.test_utils import (
    ClassicAPISimpleDeviceClass,
    ClassicAPISimpleDeviceImpl,
    attr_data_format,
    attribute_numpy_typed_values,
    attribute_typed_values,
    attribute_wrong_numpy_typed,
    base_type,
    command_numpy_typed_values,
    command_typed_values,
    dev_encoded_values,
    extract_as,
    general_typed_values,
    server_green_mode,
    server_serial_model,
    state,
    wait_for_nodb_proxy_via_pid,
)

from tango._tango import _dump_cpp_coverage

__all__ = (
    "state",
    "general_typed_values",
    "command_typed_values",
    "attribute_typed_values",
    "command_numpy_typed_values",
    "attribute_numpy_typed_values",
    "attribute_wrong_numpy_typed",
    "dev_encoded_values",
    "server_green_mode",
    "server_serial_model",
    "attr_data_format",
    "extract_as",
    "base_type",
)

device_proxy_map = {
    GreenMode.Synchronous: DeviceProxy,
    GreenMode.Futures: futures_DeviceProxy,
    GreenMode.Asyncio: partial(asyncio_DeviceProxy, wait=True),
    GreenMode.Gevent: gevent_DeviceProxy,
}


def pytest_addoption(parser):
    parser.addoption(
        "--run_extra_src_tests",
        action="store_true",
        default=False,
        help="run extra tests only for source builds",
    )
    parser.addoption(
        "--write_cpp_coverage",
        action="store_true",
        default=False,
        help="write cpp coverage data during tests",
    )


def pytest_configure(config):
    config.addinivalue_line(
        "markers", "extra_src_test: mark test as only for source builds"
    )


def pytest_collection_modifyitems(config, items):
    if config.getoption("--run_extra_src_tests"):
        # --run_extra_src_tests given in cli: do not skip those tests
        return
    skip_extra_src_test = pytest.mark.skip(
        reason="need --run_extra_src_tests option to run"
    )
    for item in items:
        if "extra_src_test" in item.keywords:
            item.add_marker(skip_extra_src_test)


@pytest.hookimpl()
def pytest_sessionfinish(session):
    """Collects all tests to be run and outputs to bat script"""
    if "--collect-only" in sys.argv and "-q" in sys.argv and "nt" in os.name:
        print("Generating windows test script...")
        script_path = os.path.join(os.path.dirname(__file__), "run_tests_win.bat")

        total_tests = len(session.items)

        with open(script_path, "w") as f:
            f.write("@echo off\r\n")
            f.write("setlocal enabledelayedexpansion\r\n")
            f.write("REM This script will run all tests separately.\r\n")
            f.write(f"set TOTAL_TESTS={total_tests}\r\n")
            f.write("set CURRENT_TEST=0\r\n")
            f.write("\r\n")

            for item in session.items:
                # Escape special characters for batch file
                nodeid = item.nodeid
                nodeid_escaped = nodeid.replace("(", "^(").replace(")", "^)")

                lines = [
                    # Progress counter
                    "set /a CURRENT_TEST+=1",
                    f"echo [!CURRENT_TEST!/%TOTAL_TESTS%] Running: {nodeid_escaped}",
                    "",
                    # Run a single test
                    f'pytest -c pytest_win_config.toml "{nodeid}"',
                    "set FIRST_RUN_ERROR=!errorlevel!",
                    "",
                    # Check if test failed (exit code 1)
                    "if !FIRST_RUN_ERROR! equ 1 (",
                    # we retry ones
                    f"    echo [!CURRENT_TEST!/%TOTAL_TESTS%] FAILED - Retrying: {nodeid_escaped}",
                    f'    pytest --lf -c pytest_win_config.toml "{nodeid}"',
                    "    set RETRY_ERROR=!errorlevel!",
                    "    if !RETRY_ERROR! equ 1 (",
                    f'        echo {nodeid_escaped} >> "%~dp0failed_tests.txt"',
                    f"        echo [!CURRENT_TEST!/%TOTAL_TESTS%] FAILED after retry: {nodeid_escaped}",
                    "    ) else if !RETRY_ERROR! equ 0 (",
                    f"        echo [!CURRENT_TEST!/%TOTAL_TESTS%] PASSED on retry: {nodeid_escaped}",
                    "    ) else if !RETRY_ERROR! geq 2 if !RETRY_ERROR! leq 5 (",
                    # Abort if pytest could not execute properly
                    # From: https://docs.pytest.org/en/7.1.x/reference/exit-codes.html
                    #   Exit code 0: All tests were collected and passed successfully
                    #   Exit code 1: Tests were collected and run but some of the tests failed
                    #   Exit code 2: Test execution was interrupted by the user
                    #   Exit code 3: Internal error happened while executing tests
                    #   Exit code 4: pytest command line usage error
                    #   Exit code 5: No tests were collected
                    "        exit /b !RETRY_ERROR!",
                    "    )",
                    ") else if !FIRST_RUN_ERROR! equ 0 (",
                    f"    echo [!CURRENT_TEST!/%TOTAL_TESTS%] PASSED: {nodeid_escaped}",
                    # Abort if pytest could not execute properly (exit codes 2-5)
                    ") else if !FIRST_RUN_ERROR! geq 2 if !FIRST_RUN_ERROR! leq 5 (",
                    "    exit /b !FIRST_RUN_ERROR!",
                    ")",
                    "",
                ]
                f.writelines([f"{line}\r\n" for line in lines])


@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_makereport():
    """Produces summary.json file for quick windows test summary"""
    summary_path = "summary.json"

    outcome = yield  # Run all other pytest_runtest_makereport non wrapped hooks
    result = outcome.get_result()
    if result.when == "call" and "nt" in os.name and os.path.isfile(summary_path):
        with open(summary_path, "r+") as f:
            summary = f.read()
            try:
                summary = json.loads(summary)
            except Exception:
                summary = []
            finally:
                outcome = str(result.outcome).capitalize()
                test = {
                    "testName": result.nodeid,
                    "outcome": outcome,
                    "durationMilliseconds": result.duration,
                    "StdOut": result.capstdout,
                    "StdErr": result.capstderr,
                }
                summary.append(test)
                f.seek(0)
                f.write(json.dumps(summary))
                f.truncate()


def start_server(host, server, inst, device):
    exe = shutil.which(server)
    cmd = f"{exe} {inst} -ORBendPoint giop:tcp:{host}:0 -nodb -dlist {device}"
    proc = Popen(cmd.split(), close_fds=True)
    proc.poll()
    return proc


@pytest.fixture(
    params=GreenMode.values.values(),
    ids=str,
    scope="module",
)
def tango_test_with_green_modes(request):
    green_mode = request.param
    server = "TangoTest"
    inst = "test"
    device = "sys/tg_test/17"
    host = "127.0.0.1"
    proc = start_server(host, server, inst, device)
    proxy = wait_for_nodb_proxy_via_pid(
        proc.pid, host, device, device_proxy_map[green_mode]
    )

    yield proxy

    proc.terminate()
    # let's not wait for it to exit, that takes too long :)


@pytest.fixture(scope="module")
def tango_test():
    green_mode = GreenMode.Synchronous
    server = "TangoTest"
    inst = "test"
    device = "sys/tg_test/17"
    host = "127.0.0.1"
    proc = start_server(host, server, inst, device)
    proxy = wait_for_nodb_proxy_via_pid(
        proc.pid, host, device, device_proxy_map[green_mode]
    )

    yield proxy

    proc.terminate()


@pytest.fixture(scope="function")
def tango_test_process_device_trl_with_function_scope():
    green_mode = GreenMode.Synchronous
    server = "TangoTest"
    inst = "test"
    device = "sys/tg_test/18"
    host = "127.0.0.1"
    proc = start_server(host, server, inst, device)
    proxy = wait_for_nodb_proxy_via_pid(
        proc.pid, host, device, device_proxy_map[green_mode]
    )

    device_trl = (
        f"tango://{proxy.get_dev_host()}:{proxy.get_dev_port()}/"
        f"{proxy.dev_name()}#dbase=no"
    )
    yield proc, device_trl

    proc.terminate()


@pytest.fixture(
    params=GreenMode.values.values(),
    ids=str,
    scope="module",
)
def green_mode_device_proxy(request):
    green_mode = request.param
    return device_proxy_map[green_mode]


def run_mixed_server():
    util = Util(
        [
            "MixedServer",
            "1",
            "-ORBendPoint",
            "giop:tcp:127.0.0.1:0",
            "-nodb",
            "-dlist",
            "my/mixed/1",
        ]
    )
    util.add_class(
        ClassicAPISimpleDeviceClass,
        ClassicAPISimpleDeviceImpl,
        "ClassicAPISimpleDevice",
    )
    util.add_class("TangoTest", "TangoTest", language="c++")
    u = Util.instance()
    u.server_init()
    u.server_run()


@pytest.fixture(autouse=True)
def flush_cpp_coverage_data(request):
    """
    Flushes C++ coverage data to disk after each test execution
    when --write_cpp_coverage command line argument was passed.
    """

    # nothing on enter
    yield

    if request.config.getoption("--write_cpp_coverage"):
        _dump_cpp_coverage()


@pytest.fixture
def mixed_tango_test_server():
    process = multiprocessing.Process(target=run_mixed_server)
    process.start()

    proxy_waiter = partial(
        wait_for_nodb_proxy_via_pid,
        process.pid,
        "127.0.0.1",
        "dserver/mixedserver/1",
        device_proxy_map[GreenMode.Synchronous],
    )
    yield process, proxy_waiter

    if process.is_alive():
        process.terminate()
        process.join(timeout=3.0)  # Allow TangoTest time to stop DataGenerator
